"""Handler for clearing main story chapters"""
from typing import Any


from ... import helper
from . import story_level_id_selector

CHAPTERS = [
    "世界篇第一章",
    "世界篇第二章",
    "世界篇第三章",
    "未來篇第一章",
    "未來篇第二章",
    "未來篇第三章",
    "宇宙篇第一章",
    "宇宙篇第二章",
    "宇宙篇第三章",
]


def clear_specific_level_ids(
    save_stats: dict[str, Any], chapter_id: int, progress: int
) -> dict[str, Any]:
    """Clear specific levels in a chapter"""
    story_chapters = save_stats["story_chapters"]
    progress = helper.clamp(progress, 0, 48)
    if progress == 0:
        story_chapters["Chapter Progress"][chapter_id] = 0
        story_chapters["Times Cleared"][chapter_id] = [0] * 51
    else:
        stage_index = progress - 1
        story_chapters["Chapter Progress"][chapter_id] = progress
        # set all levels before the one being cleared to 1
        story_chapters["Times Cleared"][chapter_id][stage_index] = 1
        for i in range(stage_index):
            story_chapters["Times Cleared"][chapter_id][i] = 1
        # set all levels after the one being cleared to 0
        for i in range(stage_index + 1, get_total_stages(save_stats, chapter_id) + 3):
            story_chapters["Times Cleared"][chapter_id][i] = 0

    save_stats["story_chapters"] = story_chapters
    return save_stats


def has_cleared_chapter(save_stats: dict[str, Any], chapter_id: int) -> bool:
    """
    Check if a chapter has been cleared

    Args:
        save_stats (dict[str, Any]): Save stats
        chapter_id (int): Chapter ID

    Returns:
        bool: True if cleared, False if not
    """
    chapter_id = format_story_id(chapter_id)

    return save_stats["story_chapters"]["Chapter Progress"][chapter_id] >= 48


def format_story_ids(ids: list[int]) -> list[int]:
    """For some reason there is a gap after EoC 3. This adds that"""

    formatted_ids: list[int] = []
    for story_id in ids:
        formatted_ids.append(format_story_id(story_id))
    return formatted_ids


def format_story_id(chapter_id: int) -> int:
    """For some reason there is a gap after EoC 3. This adds that"""

    if chapter_id > 2:
        chapter_id += 1
    return chapter_id


def clear_levels(
    story_chapters: dict[str, Any],
    treasures: list[list[int]],
    ids: list[int],
    val: int,
    chapter_progress: int,
    clear: bool,
) -> tuple[dict[str, Any], list[list[int]]]:
    """Clear levels in a chapter"""

    for chapter_id in ids:
        story_chapters["Chapter Progress"][chapter_id] = chapter_progress
        story_chapters["Times Cleared"][chapter_id] = (
            ([val] * chapter_progress) + ([0] * (48 - chapter_progress)) + ([0] * 3)
        )
        if not clear:
            treasures[chapter_id] = [0] * 49
    return story_chapters, treasures


def get_total_stages(save_stats: dict[str, Any], chapter_id: int) -> int:
    """Get the total number of stages in a chapter"""

    return len(save_stats["story_chapters"]["Times Cleared"][chapter_id]) - 3


def clear_each(save_stats: dict[str, Any]):
    """Clear stages for each chapter"""

    chapter_ids = story_level_id_selector.select_specific_chapters()

    for chapter_id in chapter_ids:
        helper.colored_text(f"Chapter: &{chapter_id+1}& : &{CHAPTERS[chapter_id]}&")
        formatted_id = format_story_id(chapter_id)
        progress = story_level_id_selector.select_level_progress(
            chapter_id, get_total_stages(save_stats, formatted_id)
        )
        save_stats = clear_specific_level_ids(save_stats, formatted_id, progress)
    helper.colored_text("Successfully set main story chapters")
    return save_stats


def clear_all(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear whole chapters"""

    chapter_ids = story_level_id_selector.select_specific_chapters()
    text = ""
    for chapter_id in chapter_ids:
        text += f"Chapter: &{chapter_id+1}& : &{CHAPTERS[chapter_id]}&\n"
    helper.colored_text(text.strip("\n"))
    progress = story_level_id_selector.select_level_progress(
        None, get_total_stages(save_stats, 0)
    )
    for chapter_id in chapter_ids:
        chapter_id = format_story_id(chapter_id)
        save_stats = clear_specific_level_ids(save_stats, chapter_id, progress)
    helper.colored_text("Successfully set main story chapters")
    return save_stats



from typing import Any

from ... import item, csv_handler, game_data_getter, helper


def get_fruit_names(is_jp: bool) -> list[str]:
    """取得貓薄荷名稱"""

    file_data = game_data_getter.get_file_latest("resLocal", "GatyaitemName.csv", is_jp)
    if file_data is None:
        helper.error_text("無法取得貓薄荷名稱")
        return []
    item_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    file_data = game_data_getter.get_file_latest("DataLocal", "Matatabi.tsv", is_jp)
    if file_data is None:
        helper.error_text("無法取得 Matatabi 資料")
        return []
    fruit_ids = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
            delimeter="\t",
        )
    )[1:]
    fruit_names: list[str] = []
    for fruit in fruit_ids:
        fruit_names.append(item_names[int(fruit[0])][0])
    return fruit_names


def edit_catfruit(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯貓薄荷 / 獸石的處理器"""

    max_cf = 128
    if save_stats["game_version"]["Value"] >= 110400:
        max_cf = None

    catfruit = item.IntItemGroup.from_lists(
        names=get_fruit_names(helper.check_data_is_jp(save_stats)),
        values=save_stats["cat_fruit"],
        maxes=max_cf,
        group_name="貓薄荷 / 獸石",
    )
    catfruit.edit()
    save_stats["cat_fruit"] = catfruit.get_values()
    return save_stats




"""Handler for editing the ototo cat cannon"""
from typing import Any, Optional

from ... import user_input_handler, game_data_getter, csv_handler, helper


def get_canon_types(is_jp: bool) -> Optional[list[str]]:
    """Get the cannon types"""

    file_data = game_data_getter.get_file_latest(
        "resLocal", "CastleRecipeDescriptions.csv", is_jp
    )
    if file_data is None:
        helper.error_text("Could not find CastleRecipeDescriptions.csv")
        return None
    data = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    types: list[str] = []
    for cannon in data:
        types.append(cannon[1])
    return types


def get_cannon_maxes(is_jp: bool) -> Optional[dict[int, dict[int, int]]]:
    """Get the cannon maxes"""
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "CastleRecipeUnlock.csv", is_jp
    )
    if file_data is None:
        helper.error_text("Could not find CastleRecipeUnlock.csv")
        return None
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    maxes: dict[int, dict[int, int]] = {}
    for cannon in data:
        cannon_id = cannon[0]
        part = cannon[1]
        max_val = cannon[-1]
        if cannon_id not in maxes:
            maxes[cannon_id] = {}
        if part not in maxes[cannon_id]:
            maxes[cannon_id][part] = max_val
        elif max_val > maxes[cannon_id][part]:
            maxes[cannon_id][part] = max_val
    return maxes


def get_part_id_from_str(part: str) -> int:
    """Get the part id from the string"""
    if part == "effect":
        return 0
    if part == "foundation":
        return 1
    if part == "style":
        return 2
    return 0


def get_max(
    part: str, cannon_id: int, cannon_maxes: dict[int, dict[int, int]]
) -> Optional[int]:
    """Get the max value for the part"""
    part_id = get_part_id_from_str(part)
    if cannon_id not in cannon_maxes:
        return None
    if part_id not in cannon_maxes[cannon_id]:
        return None
    return cannon_maxes[cannon_id][part_id]


def edit_cat_cannon(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for ototo cat cannon upgrades"""

    cannons: dict[int, dict[str, Any]] = save_stats["ototo_cannon"]

    cannon_types = get_canon_types(helper.check_data_is_jp(save_stats))
    if cannon_types is None:
        return save_stats

    cannon_maxes = get_cannon_maxes(helper.check_data_is_jp(save_stats))
    if cannon_maxes is None:
        return save_stats

    extra_data: list[str] = []
    for i in range(len(cannon_types)):
        levels = cannons[i]["levels"]
        if i == 0:
            extra_data.append(f"Level: &{levels['effect']+1}&")
            continue
        string = ""
        for level_str, level in levels.items():
            part_id = get_part_id_from_str(level_str)
            if part_id == 0:
                level += 1
            string += f"{level_str.title()}: &{level}&, "
        string = string[:-2]
        string += f" (Development: &{cannons[i]['unlock_flag']}&)"
        extra_data.append(string)

    cannon_ids = user_input_handler.select_not_inc(cannon_types, extra_data=extra_data)
    if len(cannon_ids) > 1:
        individual = user_input_handler.ask_if_individual("Cat Cannons")
    else:
        individual = True

    if individual:
        for cannon_id in cannon_ids:
            helper.colored_text(
                f"Editing &{cannon_types[cannon_id]}&", helper.WHITE, helper.GREEN
            )
            cannon = cannons[cannon_id]
            if cannon_id == 0:
                max = get_max("effect", cannon_id, cannon_maxes)
                if max is None:
                    continue
                level = user_input_handler.get_int(
                    f"Enter the level to upgrade the base to (Max &{max}&):",
                )
                level -= 1
                level = helper.config_clamp(level, 0, max)
                cannon["levels"]["effect"] = level
                continue
            develop_stage = (
                user_input_handler.colored_input(
                    "Do you want to set the stage of development (&1&) or the upgrade level? (&2&):",
                )
                == "1"
            )
            if develop_stage:
                unlock_flag = user_input_handler.get_int(
                    "Enter the stage of development (1=effect, 2=foundation, 3=style):",
                )
                unlock_flag = helper.config_clamp(unlock_flag, 0, 3)
                cannon["unlock_flag"] = unlock_flag
                if unlock_flag != 3:
                    for level_str in cannon["levels"]:
                        cannon["levels"][level_str] = 0
            else:
                cannon["upgrade_flag"] = 3
                for level_str in cannon["levels"]:
                    max = get_max(level_str, cannon_id, cannon_maxes)
                    if max is None:
                        continue
                    part_id = get_part_id_from_str(level_str)
                    level = user_input_handler.get_int(
                        f"Enter the level to upgrade &{level_str}& to (Max &{max}&):"
                    )
                    if part_id == 0:
                        level -= 1
                    level = helper.config_clamp(level, 0, max)
                    cannon["levels"][level_str] = level
    else:
        develop_stage = (
            user_input_handler.colored_input(
                "Do you want to set the stage of development (&1&) or the upgrade level? (&2&):",
            )
            == "1"
        )
        if develop_stage:
            unlock_value = user_input_handler.get_int(
                "Enter the stage of development (1=effect, 2=foundation, 3=style):",
            )
            unlock_value = helper.config_clamp(unlock_value, 0, 3)
            for cannon_id in cannon_ids:
                cannons[cannon_id]["unlock_flag"] = unlock_value
                if unlock_value != 3:
                    for level_str in cannons[cannon_id]["levels"]:
                        cannons[cannon_id]["levels"][level_str] = 0
        else:
            max_max = 0
            for cannon_id in cannon_ids:
                for part_id in cannon_maxes[cannon_id]:
                    if cannon_maxes[cannon_id][part_id] > max_max:
                        max_max = cannon_maxes[cannon_id][part_id]

            level = user_input_handler.get_int(
                f"Enter the level to upgrade everything to (Max &{max_max}&):",
            )
            for cannon_id in cannon_ids:
                cannon = cannons[cannon_id]
                cannon["upgrade_flag"] = 3
                for level_str in cannon["levels"]:
                    max = get_max(level_str, cannon_id, cannon_maxes)
                    if max is None:
                        continue
                    part_id = get_part_id_from_str(level_str)
                    level_ = level
                    if part_id == 0:
                        level_ -= 1
                    if cannon_id == 0:
                        part_id = 0
                    level_ = helper.config_clamp(
                        level_, 0, cannon_maxes[cannon_id][part_id]
                    )
                    cannon["levels"][level_str] = level_

    return save_stats



"""處理故事模式關卡選擇的模組"""
from typing import Optional

from ... import user_input_handler, helper
from . import main_story


def select_specific_chapters() -> list[int]:
    """選擇特定章節"""

    print("你想選擇哪些章節？")
    ids = user_input_handler.select_not_inc(main_story.CHAPTERS, "clear")
    return ids


def get_option():
    """取得選項"""

    options = [
        "以關卡 ID 選擇特定關卡",
        "選擇直到某關為止的所有關卡",
        "選擇所有關卡",
    ]
    return user_input_handler.select_single(options)


def select_levels(
    chapter_id: Optional[int], forced_option: Optional[int] = None, total: int = 48
) -> list[int]:
    """選擇關卡"""

    if forced_option is None:
        choice = get_option()
    else:
        choice = forced_option
    if choice == 1:
        return select_specific_levels(chapter_id, total)
    if choice == 2:
        return select_levels_up_to(chapter_id, total)
    if choice == 3:
        return select_all(total)
    return []


def select_specific_levels(chapter_id: Optional[int], total: int) -> list[int]:
    """選擇特定關卡"""

    print("你想選擇哪些關卡？")
    if chapter_id is not None:
        helper.colored_text(
            f"章節：&{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    ids = user_input_handler.get_range_ids(
        "關卡 ID（例如 &1&=韓國、&2&=蒙古）", total
    )
    ids = helper.check_clamp(ids, total, 1, -1)
    return ids


def select_levels_up_to(chapter_id: Optional[int], total: int) -> list[int]:
    """選擇直到某一關為止的所有關卡"""

    print("你想選擇哪些關卡？")
    if chapter_id is not None:
        helper.colored_text(
            f"章節：&{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    stage_id = user_input_handler.get_int(
        f"請輸入要通過到哪一關（包含該關）（例如 &1&=通過韓國，&2&=通過韓國與蒙古，&{total}&=全通，&0&=全不通）?:"
    )
    stage_id = helper.clamp(stage_id, 1, total)
    return list(range(0, stage_id))


def select_all(total: int) -> list[int]:
    """選擇所有關卡"""

    return list(range(0, total))


def select_level_progress(
    chapter_id: Optional[int], total: int, examples: Optional[list[str]] = None
) -> int:
    """選擇關卡進度"""

    if examples is None:
        examples = [
            "韓國",
            "蒙古",
        ]

    print("你想通過到哪一關？（包含該關）")
    if chapter_id is not None:
        helper.colored_text(
            f"章節：&{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    progress = user_input_handler.get_int(
        f"請輸入要通過的關卡 ID（例如 &1&=通過 {examples[0]}，&2&=通過 {examples[0]} 與 {examples[1]}，&{total}&=全通，&0&=全不通）?:"
    )
    progress = helper.clamp(progress, 0, total)
    return progress



"""Handler for editting zero legends"""
from typing import Any

from . import event_stages
from ... import user_input_handler

def count_chapters(save_stats) -> int:
    data1 = save_stats.get("zero_legends", {})
    count = len(data1)
    return count
    
def count_stages(data) -> int:
    data1 = data.get("stages", {})
    count = len(data1)
    return count
    
def set_zl(stage_data, ids, lengths):
    for stage_id in ids:
        chapter_index = int(stage_id - 1)
        chapter_stages_count = count_stages(stage_data[chapter_index]["stars"][0])
        stage_data[chapter_index]["stars"][0]["stages_cleared"] = chapter_stages_count #stage count
        stage_data[chapter_index]["stars"][0]["unlock_next"] = 3 #idk what this means, but when i cleared stages myself, value was 3.
        for i in range(0, (chapter_stages_count - 1)):
            stage_data[chapter_index]["stars"][0]["stages"][i] = 1 #how many you cleared this stage
            i += 1
    return stage_data

def edit_zl(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting zero legends"""
    stage_data = save_stats["zero_legends"]
    lengths = count_chapters(save_stats)
    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter stage ids (e.g &1& = Zero Field, &2& = The Edge of Spacetime)(You can enter &all& to get all, a range e.g &1&-&8&, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths,
    )
    save_stats["zero_legends"] = set_zl(stage_data, ids, lengths)
    print("Successfully set Zero Legend Chapters.")
    return save_stats



from typing import Any, Union

from . import (
    helper,
    user_input_handler,
    config_manager,
)
from .edits import basic, cats, gamototo, levels, other, save_management


def fix_elsewhere_old(save_stats: dict[str, Any]) -> dict[str, Any]:
    """使用兩個存檔檔案修正 elsewhere 錯誤"""
    main_token = save_stats["token"]
    main_iq = save_stats["inquiry_code"]
    input("請選擇一個目前遊戲中載入且沒有 elsewhere 錯誤且未被封鎖的存檔\n按 Enter 繼續：")
    new_path = helper.select_file(
        "選擇一個乾淨的存檔檔案",
        helper.get_save_file_filetype(),
        helper.get_save_path(),
    )
    if not new_path:
        print("請選擇一個存檔檔案")
        return save_stats

    data = helper.load_save_file(new_path)
    new_stats = data["save_stats"]
    new_token = new_stats["token"]
    new_iq = new_stats["inquiry_code"]
    save_stats["token"] = new_token
    save_stats["inquiry_code"] = new_iq

    helper.colored_text(f"已替換查詢碼：&{main_iq}& 為 &{new_iq}&")
    helper.colored_text(f"已替換 Token：&{main_token}& 為 &{new_token}&")
    return save_stats


FEATURES: dict[str, Any] = {
    "存檔管理": {
        "保存存檔": save_management.save.save_save,
        "保存變更並上傳至遊戲伺服器（取得轉移碼和確認碼）": save_management.server_upload.save_and_upload,
        "保存變更至檔案": save_management.save.save,
        "保存變更並用 adb 推送存檔（不重新開啟遊戲）": save_management.save.save_and_push,
        "保存變更並用 adb 推送存檔（重新開啟遊戲）": save_management.save.save_and_push_rerun,
        "匯出存檔資料為 json": save_management.other.export,
        "用 adb 清除存檔資料（用於產生新帳號且不重新安裝遊戲）": save_management.other.clear_data,
        "上傳可封鎖物品追蹤資料（儲存或退出時自動執行）": save_management.server_upload.upload_metadata,
        "載入存檔資料": save_management.load.select,
        "轉換存檔資料版本": save_management.convert.convert_save,
    },
    "道具": {
        "貓罐頭": basic.basic_items.edit_cat_food,
        "XP": basic.basic_items.edit_xp,
        "貓咪券": {
            "銀券": basic.basic_items.edit_normal_tickets,
            "稀有券": basic.basic_items.edit_rare_tickets,
            "白金券": basic.basic_items.edit_platinum_tickets,
            "白金碎片": basic.basic_items.edit_platinum_shards,
            "傳說券": basic.basic_items.edit_legend_tickets,
        },
        "NP": basic.basic_items.edit_np,
        "統帥旗": basic.basic_items.edit_leadership,
        "戰鬥道具": basic.basic_items.edit_battle_items,
        "貓眼石": basic.catseyes.edit_catseyes,
        "貓薄荷 / 獸石": basic.catfruit.edit_catfruit,
        "本能珠珠(修復中)": basic.talent_orbs_new.edit_talent_orbs,
        "喵利達": basic.basic_items.edit_catamins,
        "物品方案（允許獲得不封鎖物品）": other.scheme_item.edit_scheme_data,
    },
    "加碼 / 小幫手": {
        "小幫手工程師": basic.basic_items.edit_engineers,
        "基礎材料": basic.ototo_base_mats.edit_base_mats,
        "喵利達": basic.basic_items.edit_catamins,
        "加碼多多經驗值 / 等級": gamototo.gamatoto_xp.edit_gamatoto_xp,
        "貓咪砲": gamototo.ototo_cat_cannon.edit_cat_cannon,
        "加碼多多隊員": gamototo.helpers.edit_helpers,
        "修復加碼導致遊戲當機": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "貓咪 / 特殊技能": {
        "取得 / 移除貓咪": {
            "取得貓咪": cats.get_remove_cats.get_cat,
            "移除貓咪": cats.get_remove_cats.remove_cats,
        },
        "升級貓咪": cats.upgrade_cats.upgrade_cats,
        "三階貓咪": {
            "取得貓咪三階": cats.evolve_cats.get_evolve,
            "移除貓咪三階": cats.evolve_cats.remove_evolve,
            "強制三階（無真形態者會成為空白貓）": cats.evolve_cats.get_evolve_forced,
        },
        "本能": {
            "為每隻選擇的貓咪單獨設定本能": cats.talents.edit_talents_individual,
            "全部選擇貓咪本能最大化 / 移除": cats.talents.max_all_talents,
        },
        "收集 / 移除貓咪指南": {
            "設定貓咪指南條目（不給予貓罐頭）": cats.clear_cat_guide.collect_cat_guide,
            "取消認領貓咪指南條目": cats.clear_cat_guide.remove_cat_guide,
        },
        "取得關卡掉落 - 移除提示視窗": cats.chara_drop.get_character_drops,
        "升級特殊技能 / 能力": cats.upgrade_blue.upgrade_blue,
    },
    "關卡 / 寶物": {
        "主線章節通關 / 未通關": {
            "為所有選擇章節通關所有關卡": levels.main_story.clear_all,
            "為每個選擇章節通關每個關卡": levels.main_story.clear_each,
        },
        "寶物": {
            "寶物組合（例如能量飲料、水晶等）": levels.treasures.treasure_groups,
            "特定關卡與章節單獨設定": levels.treasures.specific_stages,
            "特定關卡與所有章節同時設定": levels.treasures.specific_stages_all_chapters,
        },
        "殭屍關卡 / 爆發": levels.outbreaks.edit_outbreaks,
        "活動關卡": levels.event_stages.event_stages,
        "傳說故事": levels.event_stages.stories_of_legend,
        "神秘傳說": levels.uncanny.edit_uncanny,
        "零傳說": levels.zerolegends.edit_zl,
        "阿庫領域 / 門通關": levels.aku.edit_aku,
        "解鎖阿庫領域 / 門": levels.unlock_aku_realm.unlock_aku_realm,
        "試煉": levels.gauntlet.edit_gauntlet,
        "合作試煉": levels.gauntlet.edit_collab_gauntlet,
        "塔防": levels.towers.edit_tower,
        "巨獸狩獵": levels.behemoth_culling.edit_behemoth_culling,
        "未來計分挑戰": levels.itf_timed_scores.timed_scores,
        "挑戰戰鬥分數": basic.basic_items.edit_challenge_battle,
        "清除教學": levels.clear_tutorial.clear_tutorial,
        "貓咪道場分數（入門大殿）": basic.basic_items.edit_dojo_score,
        "新增謎題關卡": levels.enigma_stages.edit_enigma_stages,
        "允許反覆通關議會關卡": levels.allow_filibuster_clearing.allow_filibuster_clearing,
        "傳說任務": levels.legend_quest.edit_legend_quest,
    },
    "查詢碼 / Token / 帳號": {
        "查詢碼": basic.basic_items.edit_inquiry_code,
        "Token": basic.basic_items.edit_token,
        "修復 elsewhere 錯誤 / 解封帳號": other.fix_elsewhere.fix_elsewhere,
        "舊版修復 elsewhere 錯誤 / 解封帳號（需要兩個存檔）": fix_elsewhere_old,
        "生成新的查詢碼與 Token": other.create_new_account.create_new_account,
    },
    "其他": {
        "抽獎種子": basic.basic_items.edit_rare_gacha_seed,
        "解鎖裝備槽位": basic.basic_items.edit_unlocked_slots,
        "取得重啟包 / 回歸模式": basic.basic_items.edit_restart_pack,
        "喵勳章": other.meow_medals.medals,
        "遊玩時間": other.play_time.edit_play_time,
        "解鎖 / 移除敵人指南條目": other.unlock_enemy_guide.enemy_guide,
        "貓薄荷挑戰 / 任務": other.missions.edit_missions,
        "普通票券最大交易進度（允許不封鎖稀有票券）": other.trade_progress.set_trade_progress,
        "取得 / 移除黃金通行證": other.get_gold_pass.get_gold_pass,
        "領取 / 移除全部用戶等級獎勵（不給物品）": other.claim_user_rank_rewards.edit_rewards,
        "貓咪神社等級 / 經驗值": other.cat_shrine.edit_shrine_xp,
    },
    "修復": {
        "修復時間錯誤": other.fix_time_issues.fix_time_issues,
        "解鎖裝備選單": other.unlock_equip_menu.unlock_equip,
        "清除教學": levels.clear_tutorial.clear_tutorial,
        "修復 elsewhere 錯誤 / 解封帳號": other.fix_elsewhere.fix_elsewhere,
        "舊版修復 elsewhere 錯誤 / 解封帳號（需要兩個存檔）": fix_elsewhere_old,
        "修復加碼導致遊戲當機": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "編輯設定": {
        "編輯語言設定": config_manager.edit_locale,
        "編輯預設國碼": config_manager.edit_default_gv,
        "編輯預設存檔路徑": config_manager.edit_default_save_file_path,
        "編輯固定存檔路徑": config_manager.edit_fixed_save_path,
        "編輯編輯器設定": config_manager.edit_editor_settings,
        "編輯啟動設定": config_manager.edit_start_up_settings,
        "編輯儲存變更設定": config_manager.edit_save_changes_settings,
        "編輯伺服器設定": config_manager.edit_server_settings,
        "編輯設定檔路徑": config_manager.edit_config_path,
    },
    "離開": helper.exit_check_changes,
}


def get_feature(
    selected_features: Any, search_string: str, results: dict[str, Any]
) -> dict[str, Any]:
    """搜尋功能名稱內含搜尋字串的功能"""

    for feature in selected_features:
        feature_data = selected_features[feature]
        if isinstance(feature_data, dict):
            feature_data = get_feature(feature_data, search_string, results)
        if search_string.lower().replace(" ", "") in feature.lower().replace(" ", ""):
            results[feature] = selected_features[feature]
    return results


def show_options(
    save_stats: dict[str, Any], features_to_use: dict[str, Any]
) -> dict[str, Any]:
    """讓使用者輸入功能編號或名稱，並取得符合的功能"""

    if (
        not config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES")
        and FEATURES == features_to_use
    ):
        user_input = ""
    else:
        prompt = "你想編輯什麼？（有些選項內含子功能）"
        if config_manager.get_config_value_category(
            "EDITOR", "SHOW_FEATURE_SELECT_EXPLANATION"
        ):
            prompt += "\n你可以輸入編號以執行功能，或輸入關鍵字搜尋功能（例如輸入貓罐頭會執行貓罐頭功能，輸入票券會顯示所有修改票券的功能）\n直接按 Enter 可顯示所有功能列表"
        user_input = user_input_handler.colored_input(f"{prompt}：\n")
    user_int = helper.check_int(user_input)
    results = []
    if user_int is None:
        results = get_feature(features_to_use, user_input, {})
    else:
        if user_int < 1 or user_int > len(features_to_use) + 1:
            helper.colored_text("數值超出範圍", helper.RED)
            return show_options(save_stats, features_to_use)
        if FEATURES != features_to_use:
            if user_int - 2 < 0:
                return menu(save_stats)
            results = features_to_use[list(features_to_use)[user_int - 2]]
        else:
            results = features_to_use[list(features_to_use)[user_int - 1]]
    if not isinstance(results, dict):
        save_stats_return = results(save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return
    if len(results) == 0:
        helper.colored_text("找不到符合名稱的功能。", helper.RED)
        return menu(save_stats)
    if len(results) == 1 and isinstance(list(results.values())[0], dict):
        results = results[list(results)[0]]
    if len(results) == 1:
        save_stats_return = results[list(results)[0]](save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return

    helper.colored_list(["返回上一頁"] + list(results))
    return show_options(save_stats, results)


def menu(
    save_stats: dict[str, Any], path_save: Union[str, None] = None
) -> dict[str, Any]:
    """顯示選單並讓使用者選擇想編輯的功能"""

    if path_save:
        helper.set_save_path(path_save)
    if config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES"):
        helper.colored_list(list(FEATURES))
    save_stats = show_options(save_stats, FEATURES)

    return save_stats




"""基礎道具編輯處理器"""
from typing import Any
from ... import item, managed_item


def edit_cat_food(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯貓罐頭"""

    cat_food = item.IntItem(
        name="貓罐頭",
        value=item.Int(save_stats["cat_food"]["Value"]),
        max_value=45000,
        bannable=item.Bannable(
            managed_item.ManagedItemType.CATFOOD, save_stats["inquiry_code"]
        ),
    )
    cat_food.edit()
    save_stats["cat_food"]["Value"] = cat_food.get_value()
    return save_stats


def edit_xp(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯經驗值"""

    experience = item.IntItem(
        name="XP",
        value=item.Int(save_stats["xp"]["Value"]),
        max_value=99999999,
    )
    experience.edit()
    save_stats["xp"]["Value"] = experience.get_value()
    return save_stats


def edit_normal_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯銀券"""

    normal_tickets = item.IntItem(
        name="普通票券",
        value=item.Int(save_stats["normal_tickets"]["Value"]),
        max_value=9999,
    )
    normal_tickets.edit()
    save_stats["normal_tickets"]["Value"] = normal_tickets.get_value()
    return save_stats


def edit_rare_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯稀有票券"""

    rare_tickets = item.IntItem(
        name="稀有票券",
        value=item.Int(save_stats["rare_tickets"]["Value"]),
        max_value=299,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            work_around='&請不要直接修改稀有票券，請改用「普通票券最大交易進度」功能！這樣比較安全。',
            type=managed_item.ManagedItemType.RARE_TICKET,
        ),
    )
    rare_tickets.edit()
    save_stats["rare_tickets"]["Value"] = rare_tickets.get_value()
    return save_stats


def edit_platinum_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯白金票券"""

    platinum_tickets = item.IntItem(
        name="白金票券",
        value=item.Int(save_stats["platinum_tickets"]["Value"]),
        max_value=9,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            work_around="&請不要直接修改白金票券，請改用白金碎片修改！10個白金碎片=1張白金票券，這樣比較安全。",
            type=managed_item.ManagedItemType.PLATINUM_TICKET,
        ),
    )
    platinum_tickets.edit()
    save_stats["platinum_tickets"]["Value"] = platinum_tickets.get_value()
    return save_stats


def edit_platinum_shards(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯白金碎片"""

    ticket_amount = save_stats["platinum_tickets"]["Value"]
    max_value = 99 - (ticket_amount * 10)
    platinum_shards = item.IntItem(
        name="白金碎片",
        value=item.Int(save_stats["platinum_shards"]["Value"]),
        max_value=max_value,
    )
    platinum_shards.edit()
    save_stats["platinum_shards"]["Value"] = platinum_shards.get_value()
    return save_stats


def edit_np(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯 NP"""

    nyanko_points = item.IntItem(
        name="NP",
        value=item.Int(save_stats["np"]["Value"]),
        max_value=9999,
    )
    nyanko_points.edit()
    save_stats["np"]["Value"] = nyanko_points.get_value()
    return save_stats


def edit_leadership(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯統帥旗"""

    leadership = item.IntItem(
        name="統帥旗",
        value=item.Int(save_stats["leadership"]["Value"]),
        max_value=9999,
    )
    leadership.edit()
    save_stats["leadership"]["Value"] = leadership.get_value()
    return save_stats


def edit_battle_items(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯戰鬥道具"""

    battle_items = item.IntItemGroup.from_lists(
        names=[
            "加速",
            "尋寶雷達",
            "富豪貓",
            "貓CPU",
            "貓工作",
            "狙擊貓",
        ],
        values=save_stats["battle_items"],
        maxes=9999,
        group_name="戰鬥道具",
    )
    battle_items.edit()
    save_stats["battle_items"] = battle_items.get_values()

    return save_stats


def edit_engineers(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯小幫手工程師"""

    engineers = item.IntItem(
        name="小幫手工程師",
        value=item.Int(save_stats["engineers"]["Value"]),
        max_value=5,
    )
    engineers.edit()
    save_stats["engineers"]["Value"] = engineers.get_value()
    return save_stats


def edit_catamins(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯喵力達"""

    catamins = item.IntItemGroup.from_lists(
        names=[
            "喵力達 A",
            "喵力達 B",
            "喵力達 C",
        ],
        values=save_stats["catamins"],
        maxes=9999,
        group_name="喵力達",
    )
    catamins.edit()
    save_stats["catamins"] = catamins.get_values()
    return save_stats


def edit_inquiry_code(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯詢問碼"""

    print(
        "警告：只有在知道自己在做什麼時才編輯詢問碼！錯誤編輯會導致遊戲中 elsewhere 錯誤！"
    )
    inquiry_code = item.StrItem(
        name="詢問碼",
        value=save_stats["inquiry_code"],
    )
    inquiry_code.edit()
    save_stats["inquiry_code"] = inquiry_code.get_value()
    return save_stats


def edit_rare_gacha_seed(save_stats: dict[str, Any]) -> dict[str, Any]:
    """抽獎種子"""

    rare_gacha_seed = item.IntItem(
        name="抽獎種子",
        value=item.Int(save_stats["rare_gacha_seed"]["Value"], signed=False),
        max_value=None,
    )
    rare_gacha_seed.edit()
    save_stats["rare_gacha_seed"]["Value"] = rare_gacha_seed.get_value()
    return save_stats


def edit_unlocked_slots(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯已解鎖槽位數量"""

    unlocked_slots = item.IntItem(
        name="已解鎖槽位",
        value=item.Int(save_stats["unlocked_slots"]["Value"]),
        max_value=len(save_stats["slot_names"]),
    )
    unlocked_slots.edit()
    save_stats["unlocked_slots"]["Value"] = unlocked_slots.get_value()
    return save_stats


def edit_token(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯密碼刷新令牌（Token）"""

    print(
        "警告：只有在知道自己在做什麼時才編輯 Token！錯誤編輯會導致遊戲中 elsewhere 錯誤！"
    )
    token = item.StrItem(
        name="Token",
        value=save_stats["token"],
    )
    token.edit()
    save_stats["token"] = token.get_value()
    return save_stats


def edit_restart_pack(save_stats: dict[str, Any]) -> dict[str, Any]:
    """給予重啟包"""

    save_stats["restart_pack"]["Value"] = 1
    print("成功給予重啟包")
    return save_stats


def edit_challenge_battle(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯挑戰戰鬥分數"""

    challenge_battle = item.IntItem(
        name="挑戰戰鬥",
        value=item.Int(save_stats["challenge"]["Score"]["Value"]),
        max_value=None,
    )
    challenge_battle.edit()
    save_stats["challenge"]["Score"]["Value"] = challenge_battle.get_value()
    save_stats["challenge"]["Cleared"]["Value"] = 1
    return save_stats


def edit_legend_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯傳說券"""

    legend_tickets = item.IntItem(
        name="傳說券",
        value=item.Int(save_stats["legend_tickets"]["Value"]),
        max_value=4,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            type=managed_item.ManagedItemType.LEGEND_TICKET,
        ),
    )
    legend_tickets.edit()
    save_stats["legend_tickets"]["Value"] = legend_tickets.get_value()
    return save_stats


def edit_dojo_score(save_stats: dict[str, Any]) -> dict[str, Any]:
    """編輯道場分數"""

    if not save_stats["dojo_data"]:
        save_stats["dojo_data"] = {0: {0: 0}}

    dojo_score = item.IntItem(
        name="道場分數",
        value=item.Int(save_stats["dojo_data"][0][0]),
        max_value=None,
    )
    dojo_score.edit()
    save_stats["dojo_data"][0][0] = dojo_score.get_value()
    return save_stats



import os
import subprocess
from typing import Optional
from . import helper


def get_data_path() -> str:
    """
    Get the data path

    Returns:
        str: The data path
    """
    return "/data/data/"


def get_installed_battlecats_versions() -> Optional[list[str]]:
    """
    Get a list of installed battle cats versions

    Returns:
        Optional[list[str]]: A list of installed battle cats versions
    """
    if not is_ran_as_root():
        return None
    path = get_data_path()
    if not os.path.exists(path):
        return None
    versions: list[str] = []
    for folder in os.listdir(path):
        if folder == "jp.co.ponos.battlecats":
            versions.append("jp")
        elif folder.startswith("jp.co.ponos.battlecats"):
            versions.append(folder.replace("jp.co.ponos.battlecats", ""))
    return versions


def pull_save_data(game_version: str) -> Optional[str]:
    """
    Pull save data from a game version

    Args:
        game_version (str): The game version to pull from

    Returns:
        Optional[str]: The path to the pulled save data
    """
    if not is_ran_as_root():
        return None
    package_name = "jp.co.ponos.battlecats" + game_version.replace("jp", "")
    path = get_data_path() + package_name + "/files/SAVE_DATA"
    if not os.path.exists(path):
        return None
    return path


def is_ran_as_root() -> bool:
    """
    Check if the program is ran as root

    Returns:
        bool: If the program is ran as root
    """
    if not helper.is_android():
        return False
    try:
        os.listdir(get_data_path())
    except PermissionError:
        helper.colored_text(
            "Root access is required to get installed game versions. Try adding sudo before the run command",
            base=helper.RED,
        )
        return False
    return True


def rerun_game(version: str) -> None:
    """
    Rerun the game on the device without adb

    Args:
        version (str): The game version to rerun
    """
    if not is_ran_as_root():
        return
    package_name = "jp.co.ponos.battlecats" + version.replace("jp", "")
    subprocess.run(
        f"sudo pkill -f {package_name}", capture_output=True, check=False, shell=True
    )
    subprocess.run(
        f"sudo monkey -p {package_name} -c android.intent.category.LAUNCHER 1",
        capture_output=True,
        check=False,
        shell=True,
    )



"""貓咪id查詢工具"""

from typing import Any, Optional

from ... import csv_handler, game_data_getter, helper
from ..levels import main_story, uncanny

TYPES = [
    "普通",
    "EX",
    "稀有",
    "激稀有",
    "超激稀有",
    "傳說稀有",
]


def get_level_cap_increase_amount(cat_base_level: int) -> int:
    """
    取得等級上限提升的數量

    參數:
        cat_base_level (int): 貓的基礎等級（30 = 29）

    回傳:
        int: 等級上限提升的數量
    """
    return max(0, cat_base_level - 29)


def get_unit_max_levels(is_jp: bool) -> Optional[tuple[list[int], list[int]]]:
    """
    取得所有貓的最大基礎等級和額外等級

    參數:
        is_jp (bool): 遊戲是否為日版

    回傳:
        tuple[list[int], list[int]]: 所有貓的最大基礎等級和額外等級
    """
    file_data = game_data_getter.get_file_latest("DataLocal", "unitbuy.csv", is_jp)
    if file_data is None:
        helper.error_text("無法取得 unitbuy.csv")
        return None
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    max_base_level = helper.copy_first_n(data, 50)
    max_plus_level = helper.copy_first_n(data, 51)
    return max_base_level, max_plus_level


def get_unit_max_level(
    data: tuple[list[int], list[int]], cat_id: int
) -> tuple[int, int]:
    """
    取得某隻貓的最大基礎等級與額外等級

    參數:
        data (tuple[list[int], list[int]]): 所有貓的最大基礎等級和額外等級
        cat_id (int): 貓ID

    回傳:
        tuple[int, int]: 指定貓的最大基礎等級與額外等級
    """
    try:
        return data[0][cat_id], data[1][cat_id]
    except IndexError:
        return 0, 0


def get_rarities(is_jp: bool) -> list[int]:
    """取得所有貓稀有度的ID"""

    file_data = game_data_getter.get_file_latest(
        "DataLocal", "unitbuy.csv", is_jp
    )
    if file_data is None:
        helper.error_text("無法取得 unitbuy.csv")
        return []
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    rarity_ids = helper.copy_first_n(data, 13)
    return rarity_ids


def get_rarity(rarity_ids: list[int], is_jp: bool) -> list[int]:
    """取得指定稀有度的所有貓ID"""

    rarities = get_rarities(is_jp)
    cat_ids: list[int] = []
    for rarity_id in rarity_ids:
        for i, rarity_val in enumerate(rarities):
            if int(rarity_val) == rarity_id:
                cat_ids.append(i)
    return cat_ids


def is_legend(cat_id: int) -> bool:
    """
    判斷貓是否為傳說稀有

    參數:
        cat_id (int): 貓ID

    回傳:
        bool: 是否為傳說稀有
    """
    legends = [
        24,
        25,
        130,
        172,
        268,
        323,
        352,
        383,
        426,
        437,
        462,
        464,
        532,
        554,
        568,
        613,
        622,
        653,
    ]
    if cat_id in legends:
        return True
    return False


def is_crazed(cat_id: int) -> bool:
    """
    判斷貓是否為狂亂貓

    參數:
        cat_id (int): 貓ID

    回傳:
        bool: 是否為狂亂貓
    """
    crazed = [
        91,
        92,
        93,
        94,
        95,
        96,
        97,
        98,
        99,
    ]
    if cat_id in crazed:
        return True
    return False


def get_max_cat_level_normal(save_stats: dict[str, Any]) -> int:
    """
    取得普通貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態

    回傳:
        int: 普通貓最大等級
    """
    if main_story.has_cleared_chapter(save_stats, 1):
        return 20
    return 10


def catseyes_unlocked(save_stats: dict[str, Any]) -> bool:
    """
    判斷是否已解鎖貓眼

    參數:
        save_stats (dict[str, Any]): 存檔狀態

    回傳:
        bool: 是否已解鎖
    """
    return helper.calculate_user_rank(save_stats) >= 1600


def get_max_cat_level_special(save_stats: dict[str, Any], cat_id: int) -> int:
    """
    取得特殊貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態
        cat_id (int): 貓ID

    回傳:
        int: 特殊貓最大等級
    """
    legend = is_legend(cat_id)
    acient_curse_clear = uncanny.is_ancient_curse_clear(save_stats)
    user_rank = helper.calculate_user_rank(save_stats)
    catseyes = catseyes_unlocked(save_stats)
    eoc_cleared_2 = main_story.has_cleared_chapter(save_stats, 1)

    if not eoc_cleared_2:
        return 10
    if user_rank < 1600:
        return 20
    if not catseyes:
        return 30
    if not acient_curse_clear and not legend:
        return 40
    if not acient_curse_clear and legend:
        return 30
    if acient_curse_clear and legend:
        return 40
    return 50


def get_max_cat_level_rare(save_stats: dict[str, Any]) -> int:
    """
    取得稀有貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態

    回傳:
        int: 稀有貓最大等級
    """
    user_rank = helper.calculate_user_rank(save_stats)
    catseyes = catseyes_unlocked(save_stats)
    cleared_eoc_2 = main_story.has_cleared_chapter(save_stats, 1)
    acient_curse_clear = uncanny.is_ancient_curse_clear(save_stats)

    if not cleared_eoc_2:
        return 10
    if user_rank < 900:
        return 20
    if user_rank < 1200:
        return 25
    if not catseyes:
        return 30
    if not acient_curse_clear:
        return 40
    return 50


def get_max_level_super_rare(save_stats: dict[str, Any], cat_id: int) -> int:
    """
    取得超稀有貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態
        cat_id (int): 貓ID

    回傳:
        int: 超稀有貓最大等級
    """
    user_rank = helper.calculate_user_rank(save_stats)
    cleared_eoc_2 = main_story.has_cleared_chapter(save_stats, 1)
    acient_curse_clear = uncanny.is_ancient_curse_clear(save_stats)
    crazed = is_crazed(cat_id)
    catseyes = catseyes_unlocked(save_stats)

    if not cleared_eoc_2:
        return 10
    if crazed and user_rank < 3600:
        return 20
    if not crazed and user_rank < 1000:
        return 20
    if crazed and user_rank < 3650:
        return 25
    if not crazed and user_rank < 1300:
        return 25
    if not catseyes:
        return 30
    if not acient_curse_clear:
        return 40
    return 50


def get_max_level_uber_rare(save_stats: dict[str, Any]) -> int:
    """
    取得激稀有貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態

    回傳:
        int: 激稀有貓最大等級
    """
    user_rank = helper.calculate_user_rank(save_stats)
    cleared_eoc_2 = main_story.has_cleared_chapter(save_stats, 1)
    acient_curse_clear = uncanny.is_ancient_curse_clear(save_stats)
    catseyes = catseyes_unlocked(save_stats)

    if not cleared_eoc_2:
        return 10
    if user_rank < 1100:
        return 20
    if user_rank < 1400:
        return 25
    if not catseyes:
        return 30
    if not acient_curse_clear:
        return 40
    return 50


def get_max_level_legend_rare(save_stats: dict[str, Any]) -> int:
    """
    取得傳說稀有貓最大可升級等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態

    回傳:
        int: 傳說稀有貓最大等級
    """
    user_rank = helper.calculate_user_rank(save_stats)
    cleared_eoc_2 = main_story.has_cleared_chapter(save_stats, 1)
    acient_curse_clear = uncanny.is_ancient_curse_clear(save_stats)
    catseyes = catseyes_unlocked(save_stats)

    if not cleared_eoc_2:
        return 10
    if user_rank < 1110:
        return 20
    if user_rank < 1410:
        return 25
    if not catseyes:
        return 30
    if not acient_curse_clear:
        return 40
    return 50


def get_max_level(save_stats: dict[str, Any], rarity_index: int, cat_id: int) -> int:
    """
    取得貓可升級的最大等級

    參數:
        save_stats (dict[str, Any]): 存檔狀態
        rarity_index (int): 稀有度索引
        cat_id (int): 貓ID

    回傳:
        int: 最大等級
    """
    if rarity_index == 0:
        return get_max_cat_level_normal(save_stats)
    if rarity_index == 1:
        return get_max_cat_level_special(save_stats, cat_id)
    if rarity_index == 2:
        return get_max_cat_level_rare(save_stats)
    if rarity_index == 3:
        return get_max_level_super_rare(save_stats, cat_id)
    if rarity_index == 4:
        return get_max_level_uber_rare(save_stats)
    if rarity_index == 5:
        return get_max_level_legend_rare(save_stats)
    return 0



"""Handler to add and remove cats"""
from typing import Any

from ... import helper
from . import cat_id_selector


def get_cat(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler to get cats"""

    cat_ids = cat_id_selector.select_cats(save_stats, False)

    save_stats = get_cat_ids(
        save_stats=save_stats,
        val=1,
        string="gave",
        ids=cat_ids,
    )
    return save_stats


def remove_cats(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler to remove cats"""

    cat_ids = cat_id_selector.select_cats(save_stats, False)

    save_stats = get_cat_ids(
        save_stats=save_stats,
        val=0,
        string="removed",
        ids=cat_ids,
    )
    return save_stats


def get_cat_ids(
    save_stats: dict[str, Any], val: int, string: str, ids: list[int]
) -> dict[str, Any]:
    """Get specific cats by ids"""

    ids = helper.check_cat_ids(ids, save_stats)

    cats = save_stats["cats"]
    seen_cats = save_stats["gatya_seen_cats"]

    for cat_id in ids:
        cats[cat_id] = val
        seen_cats[cat_id] = val

    save_stats["cats"] = cats
    save_stats["gatya_seen_cats"] = seen_cats
    save_stats["menu_unlocks"][2] = 1
    print(f"Successfully {string} cats")
    return save_stats



"""Handler for evolving cats"""
from typing import Any

from ... import helper, csv_handler, game_data_getter
from . import cat_id_selector


def get_evolve(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for evolving cats"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=cat_ids,
        forced=False,
    )


def get_evolve_forced(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for evolving cats without the form check"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=cat_ids,
        forced=True,
    )


def remove_evolve(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for de-evolving cats"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=0,
        string="removed",
        ids=cat_ids,
        forced=True,
    )


def evolve_handler(
    save_stats: dict[str, Any], val: int, string: str, forced: bool
) -> dict[str, Any]:
    """Evolve specific cats"""

    ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(save_stats, val, string, ids, forced)


def get_evolve_data(is_jp: bool) -> list[int]:
    """Get max form of cats"""

    file_data = game_data_getter.get_file_latest(
        "DataLocal", "nyankoPictureBookData.csv", is_jp
    )
    if file_data is None:
        helper.error_text("Failed to get evolve data")
        return []
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    forms = helper.copy_first_n(data, 2)
    forms = helper.offset_list(forms, -1)
    return forms


def evolve_handler_ids(
    save_stats: dict[str, Any], val: int, string: str, ids: list[int], forced: bool
) -> dict[str, Any]:
    """Evolve specific cats by ids"""
    ids = helper.check_cat_ids(ids, save_stats)
    evolves = save_stats["unlocked_forms"]
    if not forced:
        form_data = get_evolve_data(helper.check_data_is_jp(save_stats))
        length = min([len(ids), len(form_data)])
        for i in range(length):
            try:
                evolves[ids[i]] = form_data[ids[i]]
            except IndexError:
                pass
    else:
        for cat_id in ids:
            evolves[cat_id] = val
    for cat_id, (unlocked_flag, current_flag) in enumerate(
        zip(evolves, save_stats["current_forms"])
    ):
        save_stats["current_forms"][cat_id] = max(unlocked_flag, current_flag)

    flags_evolved = [0 if form == 1 else form for form in evolves]
    save_stats["unlocked_forms"] = flags_evolved

    print(f"Successfully {string} true forms of cats")
    return save_stats



from typing import Optional, Union
from . import (
    managed_item,
    user_input_handler,
    helper,
    locale_handler,
    config_manager,
    user_info,
)


class Bannable:
    def __init__(
        self,
        type: "managed_item.ManagedItemType",
        inquiry_code: str,
        work_around: str = "",
    ):
        self.type = type
        self.inquiry_code = inquiry_code
        self.work_around = work_around


class Int:
    def __init__(self, value: Optional[int], byte_size: int = 4, signed: bool = True):
        self.value = value
        self.byte_size = byte_size
        self.signed = signed

    def get_max_value(self) -> int:
        if self.signed:
            return (2 ** (self.byte_size * 8 - 1)) - 1
        return (2 ** (self.byte_size * 8)) - 1


class IntItem:
    def __init__(
        self,
        name: str,
        value: Int,
        max_value: Optional[int],
        bannable: Optional[Bannable] = None,
        offset: int = 0,
    ):
        self.name = name
        self.__value = value
        disable_maxes = config_manager.get_config_value_category(
            "EDITOR", "DISABLE_MAXES"
        )
        self.max_value = max_value
        if disable_maxes:
            self.max_value = None
        self.bannable = bannable
        self.offset = offset
        self.locale_manager = locale_handler.LocalManager.from_config()

    def get_max_value(self) -> int:
        if self.max_value is not None:
            return self.max_value
        return self.__value.get_max_value()

    def show_ban_warning(self) -> bool:
        if self.bannable is None:
            return True
        helper.colored_text(self.locale_manager.search_key("ban_warning") % self.name)
        if self.bannable.work_around:
            helper.colored_text(self.bannable.work_around)
        return user_input_handler.get_yes_no(
            self.locale_manager.search_key("ban_warning_leave")
        )

    def edit(self) -> None:
        end = not self.show_ban_warning()
        if end:
            return
        original_value = self.__value.value
        helper.colored_text(
            self.locale_manager.search_key("current_item_value")
            % (self.name, self.get_value_off())
        )
        max_str = ""
        if self.max_value is not None:
            max_str = " " + self.locale_manager.search_key("max_str") % self.max_value
        new_value = user_input_handler.get_int(
            self.locale_manager.search_key("enter_value_text") % (self.name, max_str),
        )
        new_value -= self.offset
        new_value = helper.clamp(new_value, 0, self.get_max_value())
        self.__value.value = new_value
        helper.colored_text(
            self.locale_manager.search_key("item_value_changed")
            % (
                self.name,
                0 if original_value is None else original_value,
                self.get_value_off(),
            )
        )
        if self.bannable is not None and self.__value.value != original_value:
            new_value = self.__value.value
            if original_value is None:
                original_value = 0

            info = user_info.UserInfo(self.bannable.inquiry_code)
            info.update_item(self.bannable.type, self.__value.value - original_value)

    def get_value_off(self) -> int:
        if self.__value.value is None:
            return 0
        return self.__value.value + self.offset

    def get_value(self) -> int:
        if self.__value.value is None:
            return 0
        return self.__value.value

    def get_value_none(self) -> Optional[int]:
        return self.__value.value

    def set_value(self, value: int) -> None:
        self.__value.value = value


class IntItemGroup:
    def __init__(self, group_name: str, items: list[IntItem]):
        self.items = items
        self.locale_manager = locale_handler.LocalManager.from_config()
        self.group_name = group_name

    def get_values(self) -> list[int]:
        return [item.get_value() for item in self.items]

    def get_values_none(self) -> list[Optional[int]]:
        return [item.get_value_none() for item in self.items]

    def get_values_off(self) -> list[int]:
        return [item.get_value_off() for item in self.items]

    def all_none(self) -> bool:
        return all([item.get_value_none() is None for item in self.items])

    def get_names(self) -> list[str]:
        return [item.name for item in self.items]

    def edit(self) -> None:
        if not self.items:
            return
        ids, individual = user_input_handler.select_options(
            self.get_names(),
            self.locale_manager.search_key("select_l"),
            self.get_values_off() if not self.all_none() else None,
        )
        if individual:
            for id in ids:
                self.items[id].edit()
        else:
            max_value = self.get_max_max_value()
            offset = self.items[ids[0]].offset
            max_str = ""
            if self.items[ids[0]].max_value is not None:
                max_str = " " + self.locale_manager.search_key("max_str") % (
                    max_value + offset
                )
            new_value = user_input_handler.get_int(
                self.locale_manager.search_key("enter_value_text")
                % (self.group_name, max_str)
            )
            new_value -= offset
            entered_value = helper.clamp(new_value, 0, max_value)
            for id in ids:
                max_value = self.items[id].get_max_value()
                value = helper.clamp(new_value, 0, max_value)
                self.items[id].set_value(value)

            helper.colored_text(
                self.locale_manager.search_key("success_set")
                % (self.group_name, entered_value + offset)
            )

    def get_max_max_value(self) -> int:
        return max([item.get_max_value() for item in self.items])

    @staticmethod
    def from_lists(
        names: list[str],
        values: Optional[list[int]],
        maxes: Union[list[int], int, None],
        group_name: str,
        offset: int = 0,
    ) -> "IntItemGroup":
        items: list[IntItem] = []
        for i in range(len(names)):
            max_value = maxes[i] if isinstance(maxes, list) else maxes
            try:
                value = values[i] if values is not None else None
            except IndexError:
                value = None
            items.append(
                IntItem(
                    names[i],
                    Int(value),
                    max_value,
                    offset=offset,
                )
            )
        return IntItemGroup(group_name, items)


class StrItem:
    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value
        self.locale_manager = locale_handler.LocalManager.from_config()

    def edit(self) -> None:
        original_value = self.value
        helper.colored_text(
            self.locale_manager.search_key("current_item_value")
            % (self.name, self.value)
        )
        new_value = user_input_handler.colored_input(
            self.locale_manager.search_key("enter_value_text") % (self.name, "")
        )
        self.value = new_value
        helper.colored_text(
            self.locale_manager.search_key("item_value_changed")
            % (self.name, original_value, self.value)
        )

    def get_value(self) -> str:
        return self.value



"""Handler for serialising save data from dict"""

import base64
import struct
from typing import Any, Union

import dateutil.parser

from . import helper, parse_save


def write(
    save_data: list[int],
    number: Union[dict[str, int], int],
    length: Union[int, None] = None,
) -> list[int]:
    """Writes a little endian number to the save data"""
    if length is None and isinstance(number, dict):
        length = number["Length"]
    if isinstance(number, dict):
        number = number["Value"]
    if length is None:
        raise ValueError("Length is None")
    number = int(number)
    data = list(helper.num_to_bytes(number, length))

    save_data += data

    return save_data


def create_list_separated(data: list[int], length: int) -> list[int]:
    """Creates a list of bytes from a list of numbers"""

    lst: list[int] = []
    for item in data:
        byte_data = list(helper.num_to_bytes(item, length))
        lst += byte_data
    return lst


def create_list_double(data: list[float]) -> list[int]:
    """Creates a list of bytes from a list of doubles"""

    lst: list[int] = []
    for item in data:
        byte_data = list(struct.pack("d", item))
        lst += byte_data
    return lst


def write_length_data(
    save_data: list[int],
    data: Union[list[int], dict[str, list[int]]],
    length_bytes: int = 4,
    bytes_per_val: int = 4,
    write_length: bool = True,
    length: Union[int, None] = None,
) -> list[int]:
    """Writes a list of ints to the save data"""

    if write_length is False and length is None:
        length = len(data)
    if isinstance(data, dict):
        data = data["Value"]

    if write_length:
        if length is None:
            length = len(data)
        length_data = list(helper.num_to_bytes(length, length_bytes))
        save_data += length_data

    save_data += create_list_separated(data, bytes_per_val)

    return save_data


def write_length_doubles(
    save_data: list[int],
    data: Union[list[float], dict[str, list[float]]],
    length_bytes: int = 4,
    write_length: bool = True,
    length: Union[None, int] = None,
) -> list[int]:
    """Writes a list of doubles to the save data"""

    if write_length is False and length is None:
        length = len(data)
    if isinstance(data, dict):
        data = data["Value"]

    if write_length:
        if length is None:
            length = len(data)
        length_data = list(helper.num_to_bytes(length, length_bytes))
        save_data += length_data

    save_data += create_list_double(data)

    return save_data


def serialise_time_data_skip(
    save_data: list[int],
    time_data: str,
    time_stamp: float,
    dst_flag: bool,
    duplicate: dict[str, Any],
    dst: int = 0,
) -> list[int]:
    time = dateutil.parser.parse(time_data)
    save_data = write(save_data, time.year, 4)
    save_data = write(save_data, duplicate["yy"], 4)

    save_data = write(save_data, time.month, 4)
    save_data = write(save_data, duplicate["mm"], 4)

    save_data = write(save_data, time.day, 4)
    save_data = write(save_data, duplicate["dd"], 4)

    save_data = write_double(save_data, time_stamp)

    save_data = write(save_data, time.hour, 4)
    save_data = write(save_data, time.minute, 4)
    save_data = write(save_data, time.second, 4)

    if dst_flag:
        save_data = write(save_data, dst, 1)

    return save_data


def serialise_time_data(
    save_data: list[int], time: str, dst_flag: bool, dst: int = 0
) -> list[int]:
    time_d = dateutil.parser.parse(time)
    if dst_flag:
        save_data = write(save_data, dst, 1)

    save_data = write(save_data, time_d.year, 4)
    save_data = write(save_data, time_d.month, 4)
    save_data = write(save_data, time_d.day, 4)
    save_data = write(save_data, time_d.hour, 4)
    save_data = write(save_data, time_d.minute, 4)
    save_data = write(save_data, time_d.second, 4)

    return save_data


def serialise_equip_slots(
    save_data: list[int], equip_slots: list[list[int]]
) -> list[int]:
    save_data = write(save_data, len(equip_slots), 1)
    for slot in equip_slots:
        save_data = write_length_data(save_data, slot, 0, 4, False)
    return save_data


def serialise_main_story(
    save_data: list[int], story_chapters: dict[str, list[Any]]
) -> list[int]:
    save_data = write_length_data(
        save_data, story_chapters["Chapter Progress"], write_length=False
    )
    for chapter in story_chapters["Times Cleared"]:
        save_data = write_length_data(save_data, chapter, write_length=False)
    return save_data


def serialise_treasures(save_data: list[int], treasures: list[list[int]]) -> list[int]:
    for chapter in treasures:
        save_data = write_length_data(save_data, chapter, write_length=False)
    return save_data


def serialise_cat_upgrades(
    save_data: list[int], cat_upgrades: dict[str, list[int]]
) -> list[int]:
    data: list[int] = []
    length = len(cat_upgrades["Base"])
    for cat_id in range(length):
        data.append(cat_upgrades["Plus"][cat_id])
        data.append(cat_upgrades["Base"][cat_id])
    write_length_data(save_data, data, 4, 2, True, length)
    return save_data


def serialise_blue_upgrades(
    save_data: list[int], blue_upgrades: dict[str, list[int]]
) -> list[int]:
    data: list[int] = []
    length = len(blue_upgrades["Base"])
    for blue_id in range(length):
        data.append(blue_upgrades["Plus"][blue_id])
        data.append(blue_upgrades["Base"][blue_id])
    write_length_data(save_data, data, 4, 2, False)
    return save_data


def serialise_utf8_string(
    save_data: list[int],
    string: Union[dict[str, str], str],
    length_bytes: int = 4,
    write_length: bool = True,
    length: Union[int, None] = None,
) -> list[int]:
    """Writes a string to the save data"""

    if isinstance(string, dict):
        string = string["Value"]
    data = list(string.encode("utf-8"))

    save_data = write_length_data(
        save_data, data, length_bytes, 1, write_length, length
    )
    return save_data


def serialise_event_stages_current(
    save_data: list[int], event_current: dict[str, Any]
) -> list[int]:
    unknown_val = event_current["unknown"]
    total_sub_chapters = event_current["total"] // unknown_val
    stars_per_sub_chapter = event_current["stars"]
    stages_per_sub_chapter = event_current["stages"]

    save_data = write(save_data, unknown_val, 1)
    save_data = write(save_data, total_sub_chapters, 2)
    save_data = write(save_data, stars_per_sub_chapter, 1)
    save_data = write(save_data, stages_per_sub_chapter, 1)

    for i in range(len(event_current["Clear"])):
        save_data = write_length_data(save_data, event_current["Clear"][i], 1, 1, False)

    return save_data


def flatten_list(_2d_list: Union[list[list[Any]], list[Any]]) -> list[Any]:
    flat_list: list[Any] = []
    # Iterate through the outer list
    for element in _2d_list:
        if isinstance(element, list):
            # If the element is of type list, iterate through the sublist
            for item in element:
                flat_list.append(item)
        else:
            flat_list.append(element)
    return flat_list


def serialise_event_stages(
    save_data: list[int], event_stages: dict[str, Any]
) -> list[int]:
    lengths = event_stages["Lengths"]
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    for chapter in event_stages["Value"]["clear_progress"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)

    clear_amount = [0] * total * stars * stages
    clear_amount_data = event_stages["Value"]["clear_amount"]
    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                clear_amount[i * stages * stars + j * stars + k] = clear_amount_data[i][
                    k
                ][j]

    save_data = write_length_data(save_data, clear_amount, 4, 2, False)

    for chapter in event_stages["Value"]["unlock_next"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)
    return save_data


def serialse_purchase_receipts(save_data: list[int], data: dict[Any, Any]) -> list[int]:
    save_data = write(save_data, len(data), 4)
    for item in data:
        save_data = write(save_data, item["unknown_4"], 4)
        save_data = write(save_data, len(item["item_packs"]), 4)
        for string_dict in item["item_packs"]:
            save_data = serialise_utf8_string(save_data, string_dict["Value"])
            save_data = write(save_data, string_dict["unknown_1"], 1)
    return save_data


def serialise_dumped_data(
    save_data: list[int], data: list[dict[str, int]]
) -> list[int]:
    for item in data:
        save_data = write(save_data, item)
    return save_data


def serialise_outbreaks(save_data: list[int], outbreaks: dict[Any, Any]) -> list[int]:
    save_data = write(save_data, len(outbreaks), 4)
    for chapter_id in outbreaks:
        save_data = write(save_data, int(chapter_id), 4)
        save_data = write(save_data, len(outbreaks[chapter_id]), 4)
        for level_id in outbreaks[chapter_id]:
            save_data = write(save_data, level_id, 4)
            save_data = write(save_data, outbreaks[chapter_id][level_id], 1)

    return save_data


def serialise_ototo_cat_cannon(
    save_data: list[int], ototo_cannon: dict[int, Any]
) -> list[int]:
    save_data = write(save_data, len(ototo_cannon), 4)
    for cannon_id in ototo_cannon:
        cannon = ototo_cannon[cannon_id]

        save_data = write(save_data, int(cannon_id), 4)
        save_data = write(save_data, cannon["len_val"], 4)
        save_data = write(save_data, cannon["unlock_flag"], 4)
        levels = cannon["levels"]
        save_data = write(save_data, levels["effect"], 4)
        if cannon["len_val"] == 4:
            save_data = write(save_data, levels["foundation"], 4)
            save_data = write(save_data, levels["style"], 4)
    return save_data


def serialise_uncanny_current(
    save_data: list[int], uncanny_current: dict[str, Any]
) -> list[int]:
    total_sub_chapters = uncanny_current["total"]
    stars_per_sub_chapter = uncanny_current["stars"]
    stages_per_sub_chapter = uncanny_current["stages"]

    save_data = write(save_data, total_sub_chapters, 4)
    save_data = write(save_data, stages_per_sub_chapter, 4)
    save_data = write(save_data, stars_per_sub_chapter, 4)

    for i in range(len(uncanny_current["Clear"])):
        save_data = write_length_data(
            save_data, uncanny_current["Clear"][i], 4, 4, False
        )

    return save_data


def serialise_event_timed_scores(
    save_data: list[int], timed_scores: dict[str, Any]
) -> list[int]:
    total_sub_chapters = timed_scores["total"]
    stars_per_sub_chapter = timed_scores["stars"]
    stages_per_sub_chapter = timed_scores["stages"]

    save_data = write(save_data, total_sub_chapters, 4)
    save_data = write(save_data, stages_per_sub_chapter, 4)
    save_data = write(save_data, stars_per_sub_chapter, 4)

    for i in range(len(timed_scores["Score"])):
        save_data = write_length_data(save_data, timed_scores["Score"][i], 4, 4, False)

    return save_data


def serialise_uncanny_progress(
    save_data: list[int], uncanny: dict[str, Any]
) -> list[int]:
    lengths = uncanny["Lengths"]
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    for chapter in uncanny["Value"]["clear_progress"]:
        save_data = write_length_data(save_data, chapter, 4, 4, False)

    clear_amount = [0] * total * stars * stages
    clear_amount_data = uncanny["Value"]["clear_amount"]
    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                clear_amount[i * stages * stars + j * stars + k] = clear_amount_data[i][
                    k
                ][j]

    save_data = write_length_data(save_data, clear_amount, 4, 4, False)

    for chapter in uncanny["Value"]["unlock_next"]:
        save_data = write_length_data(save_data, chapter, 4, 4, False)
    return save_data


def serialise_talent_data(save_data: list[int], talents: dict[str, Any]) -> list[int]:
    save_data = write(save_data, len(talents), 4)
    for cat_id in talents:
        cat_talent_data = talents[cat_id]
        save_data = write(save_data, int(cat_id), 4)
        save_data = write(save_data, len(cat_talent_data), 4)
        for talent in cat_talent_data:
            save_data = write(save_data, talent["id"], 4)
            save_data = write(save_data, talent["level"], 4)
    return save_data


def serialise_gauntlet_current(
    save_data: list[int], gauntlet_current: dict[str, Any]
) -> list[int]:
    save_data = write(save_data, gauntlet_current["total"], 2)
    save_data = write(save_data, gauntlet_current["stages"], 1)
    save_data = write(save_data, gauntlet_current["stars"], 1)

    for i in range(len(gauntlet_current["Clear"])):
        save_data = write_length_data(
            save_data, gauntlet_current["Clear"][i], 1, 1, False
        )

    return save_data


def serialise_gauntlet_progress(
    save_data: list[int], gauntlets: dict[str, Any]
) -> list[int]:
    lengths = gauntlets["Lengths"]
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    for chapter in gauntlets["Value"]["clear_progress"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)

    clear_amount = [0] * total * stars * stages
    clear_amount_data = gauntlets["Value"]["clear_amount"]
    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                clear_amount[i * stages * stars + j * stars + k] = clear_amount_data[i][
                    k
                ][j]

    save_data = write_length_data(save_data, clear_amount, 4, 2, False)

    for chapter in gauntlets["Value"]["unlock_next"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)
    return save_data


def serialise_legend_quest_current(
    save_data: list[int], legend_quest_current: dict[str, Any]
) -> list[int]:
    save_data = write(save_data, legend_quest_current["total"], 1)
    save_data = write(save_data, legend_quest_current["stages"], 1)
    save_data = write(save_data, legend_quest_current["stars"], 1)

    for i in range(len(legend_quest_current["Clear"])):
        save_data = write_length_data(
            save_data, legend_quest_current["Clear"][i], 1, 1, False
        )

    return save_data


def serialise_legend_quest_progress(
    save_data: list[int], legend_quests: dict[str, Any]
) -> list[int]:
    lengths = legend_quests["Lengths"]
    total = lengths["total"]
    stars = lengths["stars"]
    stages = lengths["stages"]

    for chapter in legend_quests["Value"]["clear_progress"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)

    clear_amount = [0] * total * stars * stages
    clear_amount_data = legend_quests["Value"]["clear_amount"]
    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                clear_amount[i * stages * stars + j * stars + k] = clear_amount_data[i][
                    k
                ][j]

    tries = [0] * total * stars * stages
    tries_data = legend_quests["Value"]["tries"]
    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                tries[i * stages * stars + j * stars + k] = tries_data[i][k][j]

    save_data = write_length_data(save_data, clear_amount, 4, 2, False)
    save_data = write_length_data(save_data, tries, 4, 2, False)

    for chapter in legend_quests["Value"]["unlock_next"]:
        save_data = write_length_data(save_data, chapter, 1, 1, False)
    return save_data


def serialise_talent_orbs(
    save_data: list[int], talent_orbs: dict[str, int], game_verison: dict[str, int]
) -> list[int]:
    save_data = write(save_data, len(talent_orbs), 2)
    for orb_id in talent_orbs:
        save_data = write(save_data, int(orb_id), 2)
        if game_verison["Value"] < 110400:
            save_data = write(save_data, talent_orbs[orb_id], 1)
        else:
            save_data = write(save_data, talent_orbs[orb_id], 2)
    return save_data


def serialise_aku(save_data: list[int], aku: dict[str, Any]) -> list[int]:
    lengths = aku["Lengths"]
    save_data = write(save_data, lengths["total"], 2)
    save_data = write(save_data, lengths["stages"], 1)
    save_data = write(save_data, lengths["stars"], 1)
    save_data = serialise_gauntlet_progress(save_data, aku)
    return save_data


def serialise_tower(save_data: list[int], tower: dict[str, Any]) -> list[int]:
    save_data = write(save_data, tower["current"]["total"], 4)
    save_data = write(save_data, tower["current"]["stars"], 4)

    for i in range(len(tower["current"]["selected"])):
        save_data = write_length_data(
            save_data, tower["current"]["selected"][i], 4, 4, False
        )

    save_data = write(save_data, tower["progress"]["total"], 4)
    save_data = write(save_data, tower["progress"]["stars"], 4)

    for i in range(len(tower["progress"]["clear_progress"])):
        save_data = write_length_data(
            save_data, tower["progress"]["clear_progress"][i], 4, 4, False
        )

    total = tower["progress"]["total"]
    stages = tower["progress"]["stages"]
    stars = tower["progress"]["stars"]
    save_data = write(save_data, total, 4)
    save_data = write(save_data, stages, 4)
    save_data = write(save_data, stars, 4)

    clear_amount = [0] * total * stars * stages
    clear_amount_data = tower["progress"]["clear_amount"]

    for i in range(total):
        for j in range(stages):
            for k in range(stars):
                clear_amount[i * stages * stars + j * stars + k] = clear_amount_data[i][
                    k
                ][j]

    save_data = write_length_data(save_data, clear_amount, 4, 4, False)

    save_data = serialise_dumped_data(save_data, tower["data"])

    return save_data


def exit_serialiser(save_data: list[int], save_stats: dict[str, Any]) -> list[int]:
    return serialise_utf8_string(save_data, save_stats["hash"], write_length=False)


def check_gv(
    save_data: list[int], save_stats: dict[str, Any], game_version: int
) -> dict[str, Any]:
    if save_stats["game_version"]["Value"] < game_version:
        save_data = exit_serialiser(save_data, save_stats)
        return {"save_data": save_data, "exit": True}
    else:
        return {"save_data": save_data, "exit": False}


def serialise_medals(save_data: list[int], medals: dict[str, Any]) -> list[int]:
    save_data = write_length_data(save_data, medals["medal_data_1"], 2, 2)
    medal_data_2 = medals["medal_data_2"]
    save_data = write(save_data, len(medal_data_2), 2)
    for medal_id in medal_data_2:
        save_data = write(save_data, medal_id, 2)
        save_data = write(save_data, medal_data_2[medal_id], 1)
    return save_data


def serialise_play_time(save_data: list[int], play_time: dict[str, Any]) -> list[int]:
    frames = helper.time_to_frames(play_time)
    save_data = write(save_data, frames, 4)
    return save_data


def serialise_mission_segment(save_data: list[int], data: dict[int, Any]) -> list[int]:
    save_data = write(save_data, len(data), 4)
    for mission in data:
        save_data = write(save_data, mission, 4)
        save_data = write(save_data, data[mission], 4)
    return save_data


def serialise_missions(
    save_data: list[int], missions_data: dict[str, Any]
) -> list[int]:
    save_data = serialise_mission_segment(save_data, missions_data["states"])
    save_data = serialise_mission_segment(save_data, missions_data["requirements"])
    save_data = serialise_mission_segment(save_data, missions_data["clear_types"])
    save_data = serialise_mission_segment(save_data, missions_data["gamatoto"])
    save_data = serialise_mission_segment(save_data, missions_data["nyancombo"])
    save_data = serialise_mission_segment(save_data, missions_data["user_rank"])
    save_data = serialise_mission_segment(save_data, missions_data["expiry"])
    save_data = serialise_mission_segment(save_data, missions_data["preparing"])

    return save_data


def serialise_dojo(save_data: list[int], dojo_data: dict[int, Any]) -> list[int]:
    save_data = write(save_data, len(dojo_data), 4)
    for subchapter_id in dojo_data:
        subchapter_data = dojo_data[subchapter_id]

        save_data = write(save_data, subchapter_id, 4)
        save_data = write(save_data, len(subchapter_data), 4)

        for stage_id in subchapter_data:
            score = subchapter_data[stage_id]

            save_data = write(save_data, stage_id, 4)
            save_data = write(save_data, score, 4)
    return save_data


def write_double(save_data: list[int], number: float) -> list[int]:
    """Writes a double to the save data"""

    if isinstance(number, dict):
        number = number["Value"]
    number = float(number)
    data = struct.pack("d", number)

    save_data += data

    return save_data


def start_serialize(save_stats: dict[str, Any]) -> bytes:
    """Starts the serialisation process"""

    try:
        save_data = serialize_save(save_stats)
    except Exception as e:  # pylint: disable=broad-except
        helper.colored_text(
            "\nError: An error has occurred while serializing your save data:",
            base=helper.RED,
        )
        game_version = save_stats["game_version"]["Value"]
        if game_version < 110000:
            helper.colored_text(
                f"\nThis save is from before &11.0.0& (current save version is &{helper.gv_to_str(game_version)}&), so this is likely the cause for the issue. &The save editor is not designed to work with saves from before 11.0.0&"
            )
        raise e
    return save_data


def serialise_gold_pass(save_data: list[int], gold_pass: dict[str, Any]) -> list[int]:
    """Serialises the gold pass data"""

    save_data = write(save_data, gold_pass["officer_id"])
    save_data = write(save_data, gold_pass["renewal_times"])
    save_data = write_double(save_data, gold_pass["start_date"])
    save_data = write_double(save_data, gold_pass["expiry_date"])
    save_data = write_length_doubles(
        save_data, gold_pass["unknown_2"], write_length=False
    )
    save_data = write_double(save_data, gold_pass["start_date_2"])
    save_data = write_double(save_data, gold_pass["expiry_date_2"])
    save_data = write_double(save_data, gold_pass["unknown_3"])
    save_data = write(save_data, gold_pass["flag_2"])
    save_data = write_double(save_data, gold_pass["expiry_date_3"])

    save_data = write(save_data, len(gold_pass["claimed_rewards"]), 4)
    for item_id, amount in gold_pass["claimed_rewards"].items():
        save_data = write(save_data, item_id, 4)
        save_data = write(save_data, amount, 4)

    save_data = write(save_data, gold_pass["unknown_4"])
    save_data = write(save_data, gold_pass["unknown_5"])
    save_data = write(save_data, gold_pass["unknown_6"])

    return save_data


def serialise_unlock_popups(
    save_data: list[int],
    unlock_popups: list[tuple[int, int]],
    unknown_118: dict[str, int],
):
    """Serialises the unlock popups"""

    save_data = write(save_data, len(unlock_popups), 4)
    save_data = write(save_data, unknown_118)
    for popup_id in unlock_popups:
        save_data = write(save_data, popup_id[1], 1)
        save_data = write(save_data, popup_id[0], 4)
    return save_data


def serialise_cleared_slots(
    save_data: list[int], cleared_slots: dict[str, Any]
) -> list[int]:
    """
    Serialises the cleared slots

    Args:
        save_data (list[int]): The save data
        cleared_slots (dict[str, Any]): The cleared slots

    Returns:
        list[int]: The save data
    """
    cleared_slot_data = parse_save.ClearedSlots.from_dict(cleared_slots)
    save_data = write(save_data, len(cleared_slot_data.slots), 2)
    for slot in cleared_slot_data.slots:
        save_data = write(save_data, slot.slot_index, 2)
        for cat in slot.cats:
            save_data = write(save_data, cat.cat_id, 2)
            save_data = write(save_data, cat.cat_form, 1)
        save_data = write(save_data, slot.separator, 3)
    save_data = write(save_data, cleared_slot_data.end_index, 2)

    for stages_slot in cleared_slot_data.slot_stages:
        save_data = write(save_data, stages_slot.slot_index, 2)
        save_data = write(save_data, len(stages_slot.stages), 2)
        for stage in stages_slot.stages:
            save_data = write(save_data, stage.stage_id, 4)
    return save_data


def serialise_enigma_data(
    save_data: list[int], enigma_data: dict[str, Any], game_version: int
):
    """
    Serialises the enigma data

    Args:
        save_data (list[int]): The save data
        enigma_data (dict[str, Any]): The enigma data
    """
    save_data = write(save_data, enigma_data["energy_since_1"], 4)
    save_data = write(save_data, enigma_data["energy_since_2"], 4)
    save_data = write(save_data, enigma_data["enigma_level"], 1)
    save_data = write(save_data, enigma_data["unknown_2"], 1)
    save_data = write(save_data, enigma_data["unknown_3"], 1)

    save_data = write(save_data, len(enigma_data["stages"]), 1)
    for stage in enigma_data["stages"]:
        save_data = write(save_data, stage["level"], 4)
        save_data = write(save_data, stage["stage_id"], 4)
        save_data = write(save_data, stage["decoding_status"], 1)
        save_data = write_double(save_data, stage["start_time"])

    if game_version >= 140500:
        extra_data = enigma_data.get("extra_data")
        if extra_data is not None:
            save_data = write(save_data, 1, 1)

            save_data = write(save_data, extra_data[0], 4)
            save_data = write(save_data, extra_data[1], 4)
            save_data = write(save_data, extra_data[2], 1)
            save_data = write_double(save_data, extra_data[3])
        else:
            save_data = write(save_data, 0, 1)

    return save_data


def serialise_cat_shrine(
    save_data: list[int], shrine_data: dict[str, Any]
) -> list[int]:
    """
    Serialises the cat shrine data

    Args:
        save_data (list[int]): The save data
        shrine_data (dict[str, Any]): The shrine data

    Returns:
        list[int]: The save data
    """
    save_data = write_double(save_data, shrine_data["stamp_1"])
    save_data = write_double(save_data, shrine_data["stamp_2"])
    save_data = write(save_data, shrine_data["shrine_gone"], 1)
    save_data = write_length_data(save_data, shrine_data["flags"], 1, 1)
    save_data = write(save_data, shrine_data["xp_offering"], 4)
    return save_data


def write_variable_length_int(save_data: list[int], i: int) -> list[int]:
    """
    Writes a variable length integer to the save data (I have no idea how this works and what this does)

    Args:
        save_data (list[int]): The save data
        i (int): The integer to write

    Returns:
        list[int]: The save data
    """
    i_2 = 0
    i_3 = 0
    i = int(i)
    while i >= 128:
        i_2 |= ((i & 127) | 32768) << (i_3 * 8)
        i_3 += 1
        i >>= 7
    i_4 = i_2 | (i << (i_3 * 8))
    i_5 = i_3 + 1
    for i_6 in range(i_5):
        i_7 = (i_4 >> (((i_5 - i_6) - 1) * 8)) & 255
        save_data = write(save_data, i_7, 1)
    return save_data


def set_variable_data(
    save_data: list[int], data: tuple[dict[int, int], dict[int, int]]
) -> list[int]:
    """
    Sets the variable data

    Args:
        save_data (list[int]): The save data
        data (tuple[dict[int, int], dict[int, int]]): The variable data

    Returns:
        list[int]: The save data
    """
    save_data = write_variable_length_int(save_data, len(data[0]))
    for key, value in data[0].items():
        save_data = write_variable_length_int(save_data, key)
        save_data = write_variable_length_int(save_data, value)
    save_data = write_variable_length_int(save_data, len(data[1]))
    for key, value in data[1].items():
        save_data = write_variable_length_int(save_data, key)
        save_data = write(save_data, value, 1)
    return save_data


def serialise_login_bonuses(save_data: list[int], login_bonuses: dict[int, int]):
    """
    Serialises the login bonuses

    Args:
        save_data (list[int]): The save data
        login_bonuses (dict[int, int]): The login bonuses
    """
    save_data = write(save_data, len(login_bonuses), 4)
    for key, value in login_bonuses.items():
        save_data = write(save_data, key, 4)
        save_data = write(save_data, value, 4)
    return save_data


def serialise_tower_item_obtained(save_data: list[int], data: list[list[bool]]):
    """
    Serialises the tower item obtained data

    Args:
        save_data (list[int]): The save data
        data (list[list[bool]]): The tower item obtained data
    """
    save_data = write(save_data, len(data), 4)
    save_data = write(save_data, len(data[0]), 4)
    for row in data:
        for item in row:
            save_data = write(save_data, item, 1)
    return save_data


def write_dict(save_data: list[int], data: dict[Any, Any]) -> list[int]:
    """
    Writes a dictionary to the save data

    Args:
        save_data (list[int]): The save data
        data (dict[Any, Any]): The dictionary

    Returns:
        list[int]: The save data
    """
    save_data = write(save_data, len(data), 4)
    for key, value in data.items():
        save_data = write(save_data, key, 4)
        if isinstance(value, str):
            save_data = serialise_utf8_string(save_data, value)
        elif isinstance(value, bool):
            save_data = write(save_data, value, 1)
        else:
            save_data = write(save_data, value, 4)

    return save_data


def serialise_zero_legends(save_data: list[int], data: list[Any]):
    """
    Serialises the zero legends data

    Args:
        save_data (list[int]): The save data
        data (list[Any]): The zero legends data
    """
    save_data = write(save_data, len(data), 2)
    for chapter in data:
        unknown_1 = chapter["unknown_1"]
        save_data = write(save_data, unknown_1, 1)
        save_data = write(save_data, len(chapter["stars"]), 1)
        for star in chapter["stars"]:
            selected_stage = star["selected_stage"]
            stages_cleared = star["stages_cleared"]
            unlock_next = star["unlock_next"]
            save_data = write(save_data, selected_stage, 1)
            save_data = write(save_data, stages_cleared, 1)
            save_data = write(save_data, unlock_next, 1)
            save_data = write(save_data, len(star["stages"]), 2)
            for clear_amount in star["stages"]:
                save_data = write(save_data, clear_amount, 2)
    return save_data


def serialize_save(save_stats: dict[str, Any]) -> bytes:
    """Serialises the save stats"""

    save_data: list[int] = []

    save_data = write(save_data, save_stats["game_version"])

    save_data = write(save_data, save_stats["unknown_1"])

    save_data = write(save_data, save_stats["mute_music"])
    save_data = write(save_data, save_stats["mute_sound_effects"])

    save_data = write(save_data, save_stats["cat_food"])
    save_data = write(save_data, save_stats["current_energy"])
    if save_stats["extra_time_data"]:
        if save_stats["extra_time_data"]["Value"] != 0:
            save_data = write(save_data, save_stats["extra_time_data"])
    save_data = serialise_time_data_skip(
        save_data,
        save_stats["time"],
        save_stats["time_stamp"],
        save_stats["dst"],
        save_stats["duplicate_time"],
        save_stats["dst_val"],
    )

    save_data = write_length_data(
        save_data, save_stats["unknown_flags_1"], write_length=False
    )
    save_data = write(save_data, save_stats["upgrade_state"])

    save_data = write(save_data, save_stats["xp"])

    save_data = write(save_data, save_stats["tutorial_cleared"])

    save_data = write_length_data(
        save_data, save_stats["unknown_flags_2"], write_length=False
    )
    save_data = write(save_data, save_stats["unknown_flag_1"])

    save_data = serialise_equip_slots(save_data, save_stats["slots"])
    save_data = write(save_data, save_stats["cat_stamp_current"])

    save_data = write_length_data(
        save_data, save_stats["cat_stamp_collected"], write_length=False
    )
    save_data = write(save_data, save_stats["unknown_2"])
    save_data = write(save_data, save_stats["daily_reward_flag"])
    save_data = write_length_data(
        save_data, save_stats["unknown_116"], write_length=False
    )

    save_data = serialise_main_story(save_data, save_stats["story_chapters"])
    save_data = serialise_treasures(save_data, save_stats["treasures"])

    save_data = write_length_data(save_data, save_stats["enemy_guide"])

    save_data = write_length_data(save_data, save_stats["cats"])
    save_data = serialise_cat_upgrades(save_data, save_stats["cat_upgrades"])
    save_data = write_length_data(save_data, save_stats["current_forms"])

    save_data = serialise_blue_upgrades(save_data, save_stats["blue_upgrades"])

    save_data = write_length_data(save_data, save_stats["menu_unlocks"])
    save_data = write_length_data(save_data, save_stats["new_dialogs_1"])

    save_data = write_length_data(save_data, save_stats["battle_items"], 4, 4, False, 6)
    save_data = write_length_data(save_data, save_stats["new_dialogs_2"])

    save_data = write(save_data, save_stats["unknown_6"])
    save_data = write_length_data(
        save_data, save_stats["unknown_7"], write_length=False
    )

    save_data = write(save_data, save_stats["lock_item"])
    save_data = write_length_data(save_data, save_stats["locked_items"], 1, 1, False, 6)

    save_data = serialise_time_data(
        save_data, save_stats["second_time"], save_stats["dst"], save_stats["dst_val"]
    )

    save_data = write_length_data(
        save_data, save_stats["unknown_8"], write_length=False
    )

    save_data = serialise_time_data(
        save_data, save_stats["third_time"], save_stats["dst"], save_stats["dst_val"]
    )

    save_data = write(save_data, save_stats["unknown_9"])

    save_data = serialise_utf8_string(save_data, save_stats["thirty2_code"])

    save_data = set_variable_data(save_data, save_stats["unknown_10"])

    save_data = write_length_data(
        save_data, save_stats["unknown_11"], write_length=False
    )

    save_data = write(save_data, save_stats["normal_tickets"])
    save_data = write(save_data, save_stats["rare_tickets"])

    save_data = write_length_data(save_data, save_stats["gatya_seen_cats"])

    save_data = write_length_data(
        save_data, save_stats["unknown_12"], write_length=False
    )

    if save_stats["cat_storage"]["len"]:
        save_data = write(save_data, len(save_stats["cat_storage"]["ids"]), 2)
    save_data = write_length_data(
        save_data, save_stats["cat_storage"]["ids"], 2, 4, False
    )
    save_data = write_length_data(
        save_data, save_stats["cat_storage"]["types"], 2, 4, False
    )

    save_data = serialise_event_stages_current(save_data, save_stats["event_current"])
    save_data = serialise_event_stages(save_data, save_stats["event_stages"])

    save_data = write_length_data(
        save_data, save_stats["unknown_15"], write_length=False
    )

    save_data = write_length_data(save_data, save_stats["unit_drops"])
    save_data = write(save_data, save_stats["rare_gacha_seed"])

    save_data = write(save_data, save_stats["unknown_17"])
    save_data = write(save_data, save_stats["unknown_18"])

    save_data = serialise_time_data(
        save_data, save_stats["fourth_time"], save_stats["dst"], save_stats["dst_val"]
    )

    save_data = write_length_data(save_data, save_stats["unknown_105"], 4, 4, False)

    save_data = write_length_data(
        save_data, save_stats["unknown_107"], write_length=False, bytes_per_val=1
    )
    if save_stats["dst"]:
        save_data = serialise_utf8_string(save_data, save_stats["unknown_110"])
    unknown_108 = helper.format_text(save_stats["unknown_108"])
    save_data = write(save_data, len(unknown_108), 4)
    for i in range(len(unknown_108)):
        save_data = serialise_utf8_string(save_data, unknown_108[i])

    if save_stats["dst"]:
        save_data = write_length_doubles(
            save_data, save_stats["time_stamps"], write_length=False
        )

        save_data = write(save_data, len(save_stats["unknown_112"]), 4)
        for string in save_stats["unknown_112"]:
            save_data = serialise_utf8_string(save_data, string)
        save_data = write(save_data, save_stats["energy_notice"])
        save_data = write(save_data, save_stats["game_version_2"])

    save_data = write(save_data, save_stats["unknown_111"])
    save_data = write(save_data, save_stats["unlocked_slots"])

    save_data = write(save_data, save_stats["unknown_20"]["Length_1"], 4)
    save_data = write(save_data, save_stats["unknown_20"]["Length_2"], 4)
    save_data = write_length_data(
        save_data, save_stats["unknown_20"], write_length=False
    )

    save_data = write_length_doubles(
        save_data, save_stats["time_stamps_2"][:-1], write_length=False
    )

    save_data = write(save_data, save_stats["trade_progress"])

    if save_stats["dst"]:
        save_data = write_double(save_data, save_stats["time_stamps_2"][-1])
    else:
        save_data = write(save_data, save_stats["unknown_24"])

    save_data = serialise_cat_upgrades(save_data, save_stats["catseye_related_data"])

    save_data = write_length_data(
        save_data, save_stats["unknown_22"], write_length=False
    )

    save_data = write_length_data(save_data, save_stats["user_rank_rewards"], 4, 1)

    if not save_stats["dst"]:
        save_data = write_double(save_data, save_stats["time_stamps_2"][-1])

    save_data = write_length_data(save_data, save_stats["unlocked_forms"])

    save_data = serialise_utf8_string(save_data, save_stats["transfer_code"])
    save_data = serialise_utf8_string(save_data, save_stats["confirmation_code"])
    save_data = write(save_data, save_stats["transfer_flag"])

    lengths = save_stats["stage_data_related_1"]["Lengths"]
    length = lengths[0] * lengths[1] * lengths[2]
    save_data = write_length_data(save_data, lengths, write_length=False)
    save_data = write_length_data(
        save_data, save_stats["stage_data_related_1"], 4, 1, False, length
    )

    save_data = serialise_event_timed_scores(
        save_data, save_stats["event_timed_scores"]
    )

    save_data = serialise_utf8_string(save_data, save_stats["inquiry_code"])
    save_data = serialise_play_time(save_data, save_stats["play_time"])

    save_data = write(save_data, save_stats["unknown_25"])
    save_data = write(save_data, save_stats["backup_state"])
    if save_stats["dst"]:
        save_data = write(save_data, save_stats["unknown_119"])
    save_data = write(save_data, save_stats["gv_44"])
    save_data = write(save_data, save_stats["unknown_120"])

    save_data = write_length_data(
        save_data,
        flatten_list(save_stats["itf_timed_scores"]),
        4,
        4,
        write_length=False,
    )
    save_data = write(save_data, save_stats["unknown_27"])

    save_data = write_length_data(save_data, save_stats["cat_related_data_1"])

    save_data = write(save_data, save_stats["unknown_28"])
    save_data = write(save_data, save_stats["gv_45"])
    save_data = write(save_data, save_stats["gv_46"])
    save_data = write(save_data, save_stats["unknown_29"])
    save_data = write_length_data(save_data, save_stats["lucky_tickets_1"])
    save_data = write_length_data(save_data, save_stats["unknown_32"])

    save_data = write(save_data, save_stats["gv_47"])
    save_data = write(save_data, save_stats["gv_48"])

    if not save_stats["dst"]:
        save_data = write(save_data, save_stats["energy_notice"])

    save_data = write_double(save_data, save_stats["account_created_time_stamp"])
    save_data = write_length_data(save_data, save_stats["unknown_35"])
    save_data = write(save_data, save_stats["unknown_36"])

    save_data = write(save_data, save_stats["user_rank_popups"])

    save_data = write(save_data, save_stats["gv_49"])
    save_data = write(save_data, save_stats["gv_50"])
    save_data = write(save_data, save_stats["gv_51"])

    save_data = write_length_data(
        save_data, save_stats["cat_guide_collected"], bytes_per_val=1
    )
    save_data = write(save_data, save_stats["gv_52"])

    save_data = write_length_doubles(
        save_data, save_stats["time_stamps_3"], write_length=False
    )

    save_data = write_length_data(save_data, save_stats["cat_fruit"])

    save_data = write_length_data(save_data, save_stats["cat_related_data_3"])
    save_data = write_length_data(save_data, save_stats["catseye_cat_data"])
    save_data = write_length_data(save_data, save_stats["catseyes"])
    save_data = write_length_data(save_data, save_stats["catamins"])

    seconds = helper.time_to_seconds(save_stats["gamatoto_time_left"])
    save_data = write_double(save_data, float(seconds))
    save_data = write(save_data, save_stats["gamatoto_exclamation"])
    save_data = write(save_data, save_stats["gamatoto_xp"])
    save_data = write(save_data, save_stats["gamamtoto_destination"])
    save_data = write(save_data, save_stats["gamatoto_recon_length"])

    save_data = write(save_data, save_stats["unknown_43"])

    save_data = write(save_data, save_stats["gamatoto_complete_notification"])

    save_data = write_length_data(save_data, save_stats["unknown_44"], bytes_per_val=1)
    save_data = write_length_data(
        save_data, save_stats["unknown_45"], bytes_per_val=12 * 4
    )

    save_data = write(save_data, save_stats["gv_53"])

    save_data = write_length_data(save_data, save_stats["helpers"])

    save_data = write(save_data, save_stats["unknown_47"])

    save_data = write(save_data, save_stats["gv_54"])
    save_data = serialse_purchase_receipts(save_data, save_stats["purchases"])
    save_data = write(save_data, save_stats["gv_54"])
    save_data = write(save_data, save_stats["gamatoto_skin"])
    save_data = write(save_data, save_stats["platinum_tickets"])

    save_data = serialise_login_bonuses(save_data, save_stats["login_bonuses"])
    save_data = write(save_data, save_stats["unknown_49"])
    save_data = write_length_data(
        save_data, save_stats["announcment"], write_length=False
    )
    save_data = write(save_data, save_stats["backup_counter"])

    save_data = write_length_data(
        save_data, save_stats["unknown_131"], write_length=False
    )

    save_data = write(save_data, save_stats["gv_55"])

    save_data = write(save_data, save_stats["unknown_51"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_113"])

    save_data = serialise_dojo(save_data, save_stats["dojo_data"])
    save_data = write(save_data, save_stats["dojo_item_lock"])
    save_data = write_length_data(
        save_data, save_stats["dojo_locks"], write_length=False, bytes_per_val=1
    )

    save_data = write(save_data, save_stats["unknown_114"])

    save_data = write(save_data, save_stats["gv_58"])

    save_data = write(save_data, save_stats["unknown_115"])

    save_data = serialise_outbreaks(save_data, save_stats["outbreaks"])

    save_data = write_double(save_data, save_stats["unknown_52"])
    save_data = write_length_data(
        save_data, save_stats["item_schemes"]["to_obtain_ids"]
    )
    save_data = write_length_data(save_data, save_stats["item_schemes"]["received_ids"])

    save_data = serialise_outbreaks(save_data, save_stats["current_outbreaks"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_55"])

    save_data = write_double(save_data, save_stats["time_stamp_4"])
    save_data = write(save_data, save_stats["gv_60"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_117"])

    save_data = write(save_data, save_stats["gv_61"])

    save_data = serialise_unlock_popups(
        save_data, save_stats["unlock_popups"], save_stats["unknown_118"]
    )

    save_data = write_length_data(save_data, save_stats["base_materials"])

    save_data = write(save_data, save_stats["unknown_56"])
    save_data = write(save_data, save_stats["unknown_57"])
    save_data = write(save_data, save_stats["unknown_58"])

    save_data = write(save_data, save_stats["engineers"])
    save_data = serialise_ototo_cat_cannon(save_data, save_stats["ototo_cannon"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_59"])
    save_data = serialise_tower(save_data, save_stats["tower"])
    save_data = serialise_missions(save_data, save_stats["missions"])
    save_data = serialise_tower_item_obtained(
        save_data, save_stats["tower_item_obtained"]
    )
    save_data = serialise_dumped_data(save_data, save_stats["unknown_61"])
    save_data = write(save_data, save_stats["challenge"]["Score"])
    save_data = write(save_data, save_stats["challenge"]["Cleared"])

    save_data = write(save_data, save_stats["gv_67"])

    save_data = write_dict(save_data, save_stats["weekly_event_missions"])
    save_data = write(save_data, save_stats["won_dojo_reward"])
    save_data = write(save_data, save_stats["event_flag_update_flag"])

    save_data = write(save_data, save_stats["gv_68"])

    save_data = write_dict(save_data, save_stats["completed_one_level_in_chapter"])
    save_data = write_dict(save_data, save_stats["displayed_cleared_limit_text"])
    save_data = write_dict(save_data, save_stats["event_start_dates"])
    save_data = write_length_data(save_data, save_stats["stages_beaten_twice"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_102"])

    save_data = serialise_uncanny_current(save_data, save_stats["uncanny_current"])
    save_data = serialise_uncanny_progress(save_data, save_stats["uncanny"])

    save_data = write(save_data, save_stats["unknown_62"])
    save_data = write_length_data(
        save_data, save_stats["unknown_63"], write_length=False
    )

    save_data = serialise_uncanny_current(
        save_data, save_stats["unknown_64"]["current"]
    )
    save_data = serialise_uncanny_progress(
        save_data, save_stats["unknown_64"]["progress"]
    )

    save_data = write(save_data, save_stats["unknown_65"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_66"])

    save_data = write_length_data(
        save_data, save_stats["lucky_tickets_2"], write_length=False
    )

    save_data = write_length_data(
        save_data, save_stats["unknown_67"], write_length=False
    )

    save_data = write(save_data, save_stats["unknown_68"])

    save_data = write(save_data, save_stats["gv_77"])

    save_data = serialise_gold_pass(save_data, save_stats["gold_pass"])

    save_data = serialise_talent_data(save_data, save_stats["talents"])
    save_data = write(save_data, save_stats["np"])

    save_data = write(save_data, save_stats["unknown_70"])

    save_data = write(save_data, save_stats["gv_80000"])

    save_data = write(save_data, save_stats["unknown_71"])

    save_data = write(save_data, save_stats["leadership"])

    save_data = write(save_data, save_stats["officer_pass_cat_id"])
    save_data = write(save_data, save_stats["officer_pass_cat_form"])

    save_data = write(save_data, save_stats["gv_80200"])

    save_data = write(save_data, save_stats["filibuster_stage_id"])
    save_data = write(save_data, save_stats["filibuster_stage_enabled"])

    save_data = write(save_data, save_stats["gv_80300"])

    save_data = write_length_data(save_data, save_stats["unknown_74"])

    save_data = write(save_data, save_stats["gv_80500"])

    save_data = write_length_data(save_data, save_stats["unknown_75"], 2)

    save_data = serialise_legend_quest_current(
        save_data, save_stats["legend_quest_current"]
    )
    save_data = serialise_legend_quest_progress(save_data, save_stats["legend_quest"])
    save_data = write_length_data(
        save_data, save_stats["unknown_133"], bytes_per_val=1, write_length=False
    )
    save_data = write_length_data(
        save_data, save_stats["legend_quest_ids"], write_length=False
    )

    save_data = serialise_dumped_data(save_data, save_stats["unknown_76"])

    save_data = write(save_data, save_stats["gv_80700"])
    if save_stats["dst"]:
        if save_stats["gv_100600"]["Value"] == 100600:
            save_data = write(save_data, save_stats["unknown_104"])
            save_data = write(save_data, save_stats["gv_100600"])

    save_data = write(save_data, save_stats["restart_pack"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_101"])
    save_data = serialise_medals(save_data, save_stats["medals"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_103"])

    save_data = serialise_gauntlet_current(save_data, save_stats["gauntlet_current"])
    save_data = serialise_gauntlet_progress(save_data, save_stats["gauntlets"])

    save_data = write_length_data(
        save_data, save_stats["unknown_77"], bytes_per_val=1, write_length=False
    )

    save_data = write(save_data, save_stats["gv_90300"])

    save_data = serialise_gauntlet_current(save_data, save_stats["unknown_78"])
    save_data = serialise_gauntlet_progress(save_data, save_stats["unknown_79"])
    save_data = write_length_data(
        save_data, save_stats["unknown_80"], bytes_per_val=1, write_length=False
    )
    save_data = serialise_enigma_data(
        save_data, save_stats["enigma_data"], save_stats["game_version"]["Value"]
    )
    save_data = serialise_cleared_slots(save_data, save_stats["cleared_slot_data"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_121"])
    save_data = serialise_gauntlet_current(
        save_data, save_stats["collab_gauntlets_current"]
    )
    save_data = serialise_gauntlet_progress(save_data, save_stats["collab_gauntlets"])
    save_data = write_length_data(
        save_data, save_stats["unknown_84"], bytes_per_val=1, write_length=False
    )

    save_data = serialise_dumped_data(save_data, save_stats["unknown_85"])

    save_data = serialise_talent_orbs(
        save_data, save_stats["talent_orbs"], save_stats["game_version"]
    )

    save_data = serialise_dumped_data(save_data, save_stats["unknown_86"])

    save_data = serialise_cat_shrine(save_data, save_stats["cat_shrine"])

    save_data = write(save_data, save_stats["unknown_130"])

    save_data = write(save_data, save_stats["gv_90900"])

    if save_stats["game_version"]["Value"] >= 110600:
        save_data = write(save_data, len(save_stats["slot_names"]), 1)
    for slot_name in save_stats["slot_names"]:
        save_data = serialise_utf8_string(save_data, slot_name)

    save_data = write(save_data, save_stats["gv_91000"])

    save_data = write(save_data, save_stats["legend_tickets"])

    save_data = write_length_data(
        save_data, save_stats["unknown_87"], bytes_per_val=5, length_bytes=1
    )
    save_data = write(save_data, save_stats["unknown_88"])

    save_data = serialise_utf8_string(save_data, save_stats["token"])
    save_data = write(save_data, save_stats["unknown_89"])
    save_data = write(save_data, save_stats["unknown_90"])
    save_data = write(save_data, save_stats["unknown_91"])

    save_data = write(save_data, save_stats["gv_100000"])
    data = check_gv(save_data, save_stats, 100100)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = write(save_data, save_stats["date_int"])

    save_data = write(save_data, save_stats["gv_100100"])
    data = check_gv(save_data, save_stats, 100300)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = write_length_data(
        save_data, save_stats["unknown_93"], bytes_per_val=19, write_length=False
    )

    save_data = write(save_data, save_stats["gv_100300"])
    data = check_gv(save_data, save_stats, 100700)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_94"])
    save_data = write(save_data, save_stats["platinum_shards"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_100"])

    save_data = write(save_data, save_stats["gv_100700"])
    data = check_gv(save_data, save_stats, 100900)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_aku(save_data, save_stats["aku"])

    save_data = write(save_data, save_stats["unknown_95"])
    save_data = serialise_dumped_data(save_data, save_stats["unknown_96"])

    save_data = write(save_data, save_stats["gv_100900"])
    data = check_gv(save_data, save_stats, 101000)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = write(save_data, save_stats["unknown_97"])

    save_data = write(save_data, save_stats["gv_101000"])
    data = check_gv(save_data, save_stats, 110000)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_98"])

    save_data = write(save_data, save_stats["gv_110000"])
    data = check_gv(save_data, save_stats, 110500)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_gauntlet_current(
        save_data, save_stats["behemoth_culling_current"]
    )
    save_data = serialise_gauntlet_progress(save_data, save_stats["behemoth_culling"])
    save_data = write_length_data(
        save_data, save_stats["unknown_124"], bytes_per_val=1, write_length=False
    )

    save_data = write(save_data, save_stats["unknown_125"])

    save_data = write(save_data, save_stats["gv_110500"])
    data = check_gv(save_data, save_stats, 110600)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = write(save_data, save_stats["unknown_126"])

    save_data = write(save_data, save_stats["gv_110600"])
    data = check_gv(save_data, save_stats, 110700)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_127"])

    if save_stats["dst"]:
        save_data = write(save_data, save_stats["unknown_128"])

    save_data = write(save_data, save_stats["gv_110700"])
    data = check_gv(save_data, save_stats, 110800)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = write(save_data, save_stats["shrine_dialogs"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_129"])

    save_data = write(save_data, save_stats["dojo_3x_speed"])

    save_data = serialise_dumped_data(save_data, save_stats["unknown_132"])

    save_data = write(save_data, save_stats["gv_110800"])
    data = check_gv(save_data, save_stats, 110900)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_135"])

    save_data = write(save_data, save_stats["gv_110900"])
    data = check_gv(save_data, save_stats, 120000)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_zero_legends(save_data, save_stats["zero_legends"])

    save_data = write(save_data, save_stats["unknown_136"])

    save_data = write(save_data, save_stats["gv_120000"])
    data = check_gv(save_data, save_stats, 120100)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_137"])

    save_data = write(save_data, save_stats["gv_120100"])
    data = check_gv(save_data, save_stats, 120200)
    save_data = data["save_data"]
    if data["exit"]:
        return bytes(save_data)

    save_data = serialise_dumped_data(save_data, save_stats["unknown_138"])

    save_data = write(save_data, save_stats["gv_120200"])

    extra_data = base64.b64decode(save_stats["extra_data"])

    save_data += extra_data

    save_data = exit_serialiser(save_data, save_stats)

    return bytes(save_data)



"""Handler for claiming all user rank rewards"""

from typing import Any

from ... import helper, user_input_handler


def claim(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Claim all user rank rewards"""

    save_stats["user_rank_rewards"] = [1] * len(save_stats["user_rank_rewards"])

    helper.colored_text("Claimed all user rank rewards", helper.GREEN)

    return save_stats


def edit_rewards(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Edit all user rank rewards"""

    option = user_input_handler.select_single(
        [
            "Claim all user rank rewards",
            "Clear all user rank rewards",
            "Back",
        ],
    )
    if option == 1:
        save_stats = claim(save_stats)
    elif option == 2:
        save_stats = clear(save_stats)
    return save_stats


def clear(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear all user rank rewards"""

    save_stats["user_rank_rewards"] = [0] * len(save_stats["user_rank_rewards"])

    helper.colored_text("Unclaimed all user rank rewards", helper.GREEN)

    return save_stats



from setuptools import setup, find_packages

setup(
    name="meowmod",
    version="3.0.0.5",
    author="小孫孫",
    author_email="sun1000526@gmail.com",
    description="Battle Cats Save File Editor - 修改器套件",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "meowmod": ["files/*", "files/locale/*"],
    },
    install_requires=[
        "colored==1.4.4",
        "pyyaml==6.0.2",
        "requests==2.32.3",
        "python-dateutil"
    ],
    python_requires='>=3.7',
    entry_points={
        "console_scripts": [
            "meowmod = meowmod.__main__:main",
        ],
    },
)



import json
from typing import Any
from . import config_manager, managed_item, helper
import os


class ManagedItems:
    def __init__(self, managed_items: dict[managed_item.ManagedItemType, int]):
        self.managed_items = managed_items

    def to_dict(self) -> dict[str, int]:
        """Convert to dict"""

        return {
            item.value: self.managed_items[item]
            for item in managed_item.ManagedItemType
        }

    @staticmethod
    def from_dict(data: dict[str, int]) -> "ManagedItems":
        """Convert from dict"""

        managed_items = {
            managed_item.ManagedItemType(item): data[item] for item in data
        }
        return ManagedItems(managed_items)


class UserInfo:
    def __init__(self, inquiry_code: str):
        self.inquiry_code = inquiry_code
        self.read_user_info()

    def get_path(self) -> str:
        """Get the path to the user info"""

        app_data_folder = config_manager.get_app_data_folder()
        path = os.path.join(app_data_folder, "user_info", self.inquiry_code + ".json")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return path

    def create_empty_user_info(self):
        managed_items = {item: 0 for item in managed_item.ManagedItemType}
        managed_items = ManagedItems(managed_items)
        data = {
            "managedItems": managed_items.to_dict(),
            "password": "",
            "authToken": "",
        }
        self.write_user_info(data)

    def read_user_info(self):
        if not os.path.exists(self.get_path()):
            self.create_empty_user_info()
        data = helper.read_file_string(self.get_path())
        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            self.create_empty_user_info()
            data = helper.read_file_string(self.get_path())
            data = json.loads(data)
        self.managed_items = ManagedItems.from_dict(data["managedItems"])
        self.password = data["password"]
        self.auth_token = data["authToken"]

    def write_user_info(self, data: dict[str, Any]):
        helper.write_file_string(self.get_path(), json.dumps(data, indent=4))

    def save(self):
        data = {
            "managedItems": self.managed_items.to_dict(),
            "password": self.password,
            "authToken": self.auth_token,
        }
        self.write_user_info(data)

    def get_managed_items(self) -> ManagedItems:
        return self.managed_items

    def get_password(self) -> str:
        return self.password

    def get_auth_token(self) -> str:
        return self.auth_token

    def set_managed_items(self, managed_items: ManagedItems):
        self.managed_items = managed_items
        self.save()

    def set_password(self, password: str):
        self.password = password
        self.save()

    def set_auth_token(self, auth_token: str):
        self.auth_token = auth_token
        self.save()

    def clear_managed_items(self):
        self.managed_items = ManagedItems(
            {item: 0 for item in managed_item.ManagedItemType}
        )
        self.save()

    def get_managed_items_lst(self) -> list[managed_item.ManagedItem]:
        items: list[managed_item.ManagedItem] = []
        for item in self.managed_items.managed_items:
            value = self.managed_items.managed_items[item]
            if value > 0:
                detail_type = managed_item.DetailType.GET
            elif value < 0:
                detail_type = managed_item.DetailType.USE
                value = abs(value)
            else:
                continue
            items.append(managed_item.ManagedItem(value, detail_type, item))
        return items

    def has_managed_items(self) -> bool:
        for item in self.managed_items.managed_items:
            value = self.managed_items.managed_items[item]
            if value != 0:
                return True
        return False

    def update_item(self, item_type: managed_item.ManagedItemType, amount: int):
        self.managed_items.managed_items[item_type] += amount
        self.save()

    @staticmethod
    def clear_all_items():
        app_data_folder = config_manager.get_app_data_folder()
        path = os.path.join(app_data_folder, "user_info")
        os.makedirs(path, exist_ok=True)
        files = helper.get_files_in_dir(path)
        for file in files:
            if file.endswith(".json"):
                info = UserInfo(file.replace(".json", ""))
                info.clear_managed_items()



"""Handler for setting into the future timed scores"""
from typing import Any, Union

from ... import item
from . import main_story


def set_scores(
    scores: list[list[int]], usr_scores: list[Union[int, None]]
) -> list[list[int]]:
    """Set the scores for a stage"""
    for i, usr_score in enumerate(usr_scores):
        if usr_score is None:
            continue
        scores[i] = ([usr_score] * 48) + ([0] * 3)
    return scores


def timed_scores(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for setting into the future timed scores"""

    scores = save_stats["itf_timed_scores"]
    print("Enter the scores for the following chapters:")
    usr_scores = item.IntItemGroup.from_lists(
        names=main_story.CHAPTERS[3:6],
        values=None,
        maxes=9999,
        group_name="Into The Future Timed Scores",
    )
    usr_scores.edit()
    save_stats["itf_timed_scores"] = set_scores(scores, usr_scores.get_values_none())

    print("Successfully set timed scores")
    return save_stats



"""Handler for patching save data"""

import hashlib
from typing import Union


def get_md5_sum(data: bytes) -> str:
    """Get MD5 sum of data."""

    return hashlib.md5(data).hexdigest()


def get_save_data_sum(save_data: bytes, game_version: str) -> str:
    """Get MD5 sum of save data."""

    if game_version in ("jp", "ja"):
        game_version = ""

    salt = f"battlecats{game_version}".encode("utf-8")
    data_to_hash = salt + save_data[:-32]

    return get_md5_sum(data_to_hash)


def detect_game_version(save_data: bytes) -> Union[str, None]:
    """Detect the game version of the save file"""

    if not save_data:
        return None

    game_versions = [
        "jp",
        "en",
        "kr",
        "tw",
    ]
    try:
        curr_hash = save_data[-32:].decode("utf-8")
    except UnicodeDecodeError as err:
        raise Exception("Invalid save hash") from err

    for game_version in game_versions:
        if curr_hash == get_save_data_sum(save_data, game_version):
            return game_version
    return None


def patch_save_data(save_data: bytes, game_version: str) -> bytes:
    """Set the md5 sum of the save data"""

    save_hash = get_save_data_sum(save_data, game_version)
    save_data = save_data[:-32] + save_hash.encode("utf-8")
    return save_data
